"""Reimplementation of the CARD cell-type deconvolution method.

Extended from https://github.com/YingMa0107/CARD

Reference
---------
Ma, Ying, and Xiang Zhou. "Spatially informed cell-type deconvolution for spatial transcriptomics."
Nature Biotechnology (2022): 1-11.

"""
from itertools import chain

import numpy as np
import pandas as pd
from scipy.spatial.distance import pdist, squareform


def obj_func(trac_xxt, UtXV, UtU, VtV, mGene, nSample, b, Lambda, beta, vecOne, V, L, alpha, sigma_e2=None):
    normNMF = trac_xxt - 2.0 * np.trace(UtXV) + np.trace(UtU @ VtV)
    sigma_e2 = normNMF / (mGene * nSample) or sigma_e2
    logX = -(mGene * nSample) * 0.5 * np.log(sigma_e2) - 0.5 * (normNMF / sigma_e2)
    temp = (V.T - b @ vecOne.T) @ L @ (V - vecOne @ b.T)
    logV = -(nSample) * 0.5 * np.sum(np.log(Lambda)) - 0.5 * (np.sum(np.diag(temp) / Lambda))
    logSigmaL2 = -(alpha + 1.0) * np.sum(np.log(Lambda)) - np.sum(beta / Lambda)
    return logX + logV + logSigmaL2


def CARDref(Xinput, U, W, phi, max_iter, epsilon, V, b, sigma_e2, Lambda):
    # Make necessary deep copies
    V = V.copy()

    # Initialize some useful items
    nSample = int(Xinput.shape[1])  # number of spatial sample points
    mGene = int(Xinput.shape[0])  # number of genes in spatial deconvolution
    k = int(U.shape[1])  # number of cell type
    L = np.zeros((nSample, nSample))
    D = np.zeros((nSample, nSample))
    V_old = np.zeros((nSample, k))
    UtU = np.zeros((k, k))
    VtV = np.zeros((k, k))
    colsum_W = np.zeros((nSample, 1))
    UtX = np.zeros((k, nSample))
    XtU = np.zeros((nSample, k))
    UtXV = np.zeros((k, k))
    temp = np.zeros((k, k))
    part1 = np.zeros((nSample, k))
    part2 = np.zeros((nSample, k))
    updateV_k = np.zeros((k))
    updateV_den_k = np.zeros((k))
    vecOne = np.ones((nSample, 1))
    diag_UtU = np.zeros((k))
    logicalLogL = False
    alpha = 1.0
    beta = nSample / 2.0
    accu_L = 0.0
    trac_xxt = (Xinput * Xinput).sum()

    # Initialize values with constant matrix caculations for increasing speed
    UtX = U.T @ Xinput
    XtU = UtX.T
    UtXV = UtX @ V
    VtV = V.T @ V
    UtU = U.T @ U
    diag_UtU = np.diag(UtU)
    if W is not None:
        colsum_W = np.sum(W, axis=1)
        D = np.diag(colsum_W)  # diagnol matrix whose entries are column
        L = D - phi * W  # graph laplacian
        colsum_W = colsum_W.reshape(nSample, 1)
        accu_L = np.sum(L)

    obj_old = obj_func(trac_xxt, UtXV, UtU, VtV, mGene, nSample, b, Lambda, beta, vecOne, V, L, alpha, sigma_e2)
    V_old = V.copy()

    # Iteration starts
    iter_converge = 0
    for i in range(max_iter):
        # logV = 0.0
        Lambda = (np.diag(temp) / 2.0 + beta) / (nSample / 2.0 + alpha + 1.0)
        if W is not None:
            b = np.sum(V.T @ L, axis=1, keepdims=True) / accu_L
            part1 = sigma_e2 * (D @ V + phi * colsum_W @ b.T)
            part2 = sigma_e2 * (phi * W @ V + colsum_W @ b.T)

        for nCT in range(k):
            updateV_den_k = Lambda[nCT] * (V[:, nCT] * diag_UtU[nCT] +
                                           (V @ UtU[:, nCT] - V[:, nCT] * diag_UtU[nCT])) + part1[:, nCT]

            updateV_k = (Lambda[nCT] * XtU[:, nCT] + part2[:, nCT]) / updateV_den_k
            V[:, nCT] = V[:, nCT] * updateV_k

        UtXV = UtX @ V
        VtV = V.T @ V
        obj = obj_func(trac_xxt, UtXV, UtU, VtV, mGene, nSample, b, Lambda, beta, vecOne, V, L, alpha)

        logicalLogL = (obj > obj_old) & ((abs(obj - obj_old) * 2.0 / abs(obj + obj_old)) < epsilon)
        # TODO: setup logging and make this debug or info
        # print(f"{i=:<4}, {obj=:.5e}, {logX=:5.2e}, {logV=:5.2e}, {logSigmaL2=:5.2e}")
        if (np.isnan(obj) | (np.sqrt(np.sum((V - V_old) * (V - V_old)) / (nSample * k)) < epsilon) | logicalLogL):
            if (i > 5):  # // run at least 5 iterations
                # print(f"Exiting at {i=}")
                iter_converge = i
                break
        else:
            obj_old = obj
            V_old = V.copy()

    res = {"V": V, "sigma_e2": sigma_e2, "Lambda": Lambda, "b": b, "Obj": obj, "iter_converge": iter_converge}
    return res


class Card:
    """The CARD cell-type deconvolution model.

    Parameters
    ----------
    sc_count : pd.DataFrame
        Reference single cell RNA-seq counts data.
    sc_meta : pd.DataFrame
        Reference cell-type label information.
    spatial_count : pd.DataFrame
        Target spatial RNA-seq counts data to be deconvoluted.
    spatial_location : pd.DataFrame, optional
        Optional spatial location for the spatial transcriptomic reads.
    ct_varname : str, optional
        Name of the cell-types column.
    ct_select : str, optional
        Selected cell-types to be considered for deconvolution.
    cell_varname : str, optional
        Name of the cells column.
    sample_varname : str, optional
        Name of the samples column.
    minCountGene : int
        Minimum number of genes required.
    minCountSpot : int
        Minimum number of spots required.
    basis
        The basis parameter.
    markers
        Markers.
    """

    def __init__(self, sc_count, sc_meta, spatial_count, spatial_location=None, ct_varname=None, ct_select=None,
                 cell_varname=None, sample_varname=None, minCountGene=100, minCountSpot=5, basis=None, markers=None):
        self.sc_count = sc_count
        self.sc_meta = sc_meta
        self.spatial_count = spatial_count
        self.spatial_location = spatial_location
        self.ct_varname = ct_varname
        self.ct_select = ct_select
        self.cell_varname = cell_varname
        self.sample_varname = sample_varname
        self.minCountGene = minCountGene
        self.minCountSpot = minCountSpot
        self.basis = basis
        self.marker = markers
        self.info_parameters = {}

        if self.cell_varname is None:
            self.cell_varname = "auto_cell_index"
            self.sc_meta[self.cell_varname] = self.sc_meta.index

    def createscRef(self):
        """CreatescRef - create reference basis matrix from reference scRNA-seq."""
        countMat = self.sc_count.copy()  # cell by gene matrix
        sc_meta = self.sc_meta.copy()
        ct_varname = self.ct_varname
        sample_varname = self.sample_varname
        if sample_varname is None:
            sc_meta["sampleID"] = "Sample"
            sample_varname = "sampleID"
        sample_id = sc_meta[sample_varname].astype(str)
        ct_sample_id = sc_meta[ct_varname] + "$*$" + sample_id
        rowSums_countMat = countMat.sum(axis=1)
        sc_meta["rowSums"] = rowSums_countMat
        rowSums_countMat_Ct = sc_meta.groupby([ct_varname, sample_varname])["rowSums"].agg("sum").to_frame()
        rowSums_countMat_Ct_Wide = rowSums_countMat_Ct.pivot_table(index=sample_varname, columns=ct_varname,
                                                                   values="rowSums", aggfunc="sum")

        # create count table by sampleID and cellType
        tab = sc_meta.groupby([sample_varname, ct_varname]).size()
        tbl = tab.unstack()

        # match column and row names
        rowSums_countMat_Ct_Wide = rowSums_countMat_Ct_Wide.reindex_like(tbl)
        rowSums_countMat_Ct_Wide = rowSums_countMat_Ct_Wide.reindex(tbl.index)

        # Compute total expression count by sample and cell type
        S_JK = rowSums_countMat_Ct_Wide.div(tbl)
        S_JK = S_JK.replace(0, np.nan)
        S_JK = S_JK.replace([np.inf, -np.inf], np.nan)
        S = S_JK.mean(axis=0).to_frame().unstack().droplevel(0)
        S = S[sc_meta[ct_varname].unique()]
        countMat["ct_sample_id"] = ct_sample_id
        Theta_S_colMean = countMat.groupby(ct_sample_id).agg("mean")
        tbl_sample = countMat.groupby([ct_sample_id]).size()
        tbl_sample = tbl_sample.reindex_like(Theta_S_colMean)
        tbl_sample = tbl_sample.reindex(Theta_S_colMean.index)
        Theta_S_colSums = countMat.groupby(ct_sample_id).agg("sum")
        Theta_S = Theta_S_colSums.copy()
        Theta_S["sum"] = Theta_S_colSums.sum(axis=1)
        Theta_S = Theta_S[list(Theta_S.columns)[:-1]].div(Theta_S["sum"], axis=0)
        Theta_S = Theta_S[list(Theta_S.columns)[:-1]]
        grp = []
        for ind in Theta_S.index:
            grp.append(ind.split("$*$")[0])
        Theta_S["grp"] = grp
        Theta = Theta_S.groupby(grp).agg("mean")
        Theta = Theta.reindex(sc_meta[ct_varname].unique())
        S = S[Theta.index]
        Theta["S"] = S.iloc[0]
        basis = Theta[list(Theta.columns)[:-1]].mul(Theta["S"], axis=0)
        self.basis = basis

    def select_ct_marker(self, ict):
        Basis = self.basis.copy()
        rest = Basis[Basis.index != ict].to_numpy().mean(axis=0)
        FC = np.log(Basis[Basis.index == ict].to_numpy().mean(axis=0) + 1e-6) - np.log(rest + 1e-6)
        markers = list(Basis.columns[np.logical_and(FC > 1.25, Basis[Basis.index == ict].to_numpy().mean(axis=0) > 0)])
        return markers

    def selectInfo(self, common_gene):
        """Select Informative Genes used in the deconvolution.

        Parameters
        ----------
        common_gene : list
            Common genes between scRNAseq count data and spatial resolved transcriptomics data.

        Returns
        -------
        list
            List of informative genes.

        """
        ct_varname = self.ct_varname
        cell_varname = self.cell_varname
        Basis = self.basis.copy()
        sc_count = self.sc_count.copy()
        sc_meta = self.sc_meta.copy()
        gene1_list = list()
        for ict in Basis.index:
            gene1_list.append(self.select_ct_marker(ict))
        gene1 = set(chain(*gene1_list))
        gene1 = list(gene1)
        gene1 = [gene for gene in gene1 if gene in common_gene]  # intersect with common_gene
        counts = sc_count[gene1]
        sd_within = pd.DataFrame(columns=counts.columns)
        for ict in Basis.index:
            series = (
                counts
                .loc[list(sc_meta[sc_meta[ct_varname] == ict][cell_varname])]
                .var(axis=0)
                .divide(
                    counts
                    .loc[list(sc_meta[sc_meta[ct_varname] == ict][cell_varname])]
                    .mean(axis=0)
                )
            )
            series.name = ict
            sd_within = pd.concat((sd_within, pd.DataFrame(series).T))

        sd_within_colMean = sd_within.mean(axis=0).index.to_frame()
        genes_to_select = sd_within.mean(axis=0) < sd_within.mean(axis=0).quantile(.99)
        genes = list(sd_within_colMean[genes_to_select].index)
        return genes

    def fit(self, max_iter=100, epsilon=1e-4):
        """Fit function for model training.

        Parameters
        ----------
        max_iter : int
            Maximum number of iterations for optimization.
        epsilon : float
            Optimization threshold.

        """
        ct_select = self.ct_select
        self.createscRef()  # create basis
        Basis = self.basis.copy()
        Basis = Basis.loc[ct_select]
        spatial_count = self.spatial_count.copy()

        # select common genes
        tmp = list(set(list(spatial_count.columns)).intersection(list(Basis.columns)))
        tmp.sort()
        common_gene = list()
        for gene in tmp:
            if "mt-" not in gene:
                common_gene.append(gene)

        # Select informative genes
        common = self.selectInfo(common_gene)
        Xinput = spatial_count.copy()
        B = Basis.copy()
        Xinput = Xinput[common]
        B = B[common]
        Xinput = Xinput.sort_index(axis="columns")
        B = B.sort_index(axis="columns")
        filter1 = Xinput.sum(axis="columns") > 0
        filter2 = Xinput.sum(axis="index") > 0
        Xinput = Xinput.loc[filter1]
        Xinput = Xinput.loc[:, filter2]
        B = B.loc[:, filter2]
        # normalize count data
        rowsumvec = Xinput.sum(axis=1)
        Xinput_norm = Xinput.copy()
        Xinput["rowsumvec"] = rowsumvec
        Xinput_norm = Xinput[list(Xinput.columns)[:-1]].div(Xinput["rowsumvec"], axis=0)

        # Spatial location
        spatial_location = self.spatial_location
        if spatial_location is None:
            kernel_mat = None
        else:
            # TODO: refactor this to preprocess?
            l1 = list(spatial_location.index)
            l2 = list(spatial_count.index)
            spatial_location = spatial_location.loc[[spot for spot in l1 if spot in l2]]
            norm_cords = spatial_location[["x", "y"]]
            norm_cords[["x"]] = norm_cords[["x"]] - norm_cords[["x"]].min()
            norm_cords[["y"]] = norm_cords[["y"]] - norm_cords[["y"]].min()
            scaleFactor = norm_cords.max().max()
            norm_cords[["x"]] = norm_cords[["x"]].div(scaleFactor)
            norm_cords[["y"]] = norm_cords[["y"]].div(scaleFactor)
            ED = squareform(pdist(norm_cords, "euclidean"))
            isigma = 0.1
            kernel_mat = np.exp(-ED**2 / (2 * isigma**2))
            np.fill_diagonal(kernel_mat, 0)

        # Initialize the proportion matrix
        rng = np.random.default_rng(20200107)
        Vint1 = rng.dirichlet(np.repeat(10, B.shape[0], axis=0), Xinput_norm.shape[0])
        phi = [0.01, 0.1, 0.3, 0.5, 0.7, 0.9, 0.99]
        # scale the Xinput_norm and B to speed up the convergence.
        mean_X = Xinput_norm.values.mean()
        mean_B = B.values.mean()
        Xinput_norm = Xinput_norm * 0.1 / mean_X
        B = B * 0.1 / mean_B
        # print("B", B.values.shape, mean_B.shape)

        # Optimization
        ResList = {}
        Obj = np.array([])
        for iphi in range(len(phi)):
            res = CARDref(Xinput=Xinput_norm.T.values, U=B.T.values, W=kernel_mat, phi=phi[iphi], max_iter=max_iter,
                          epsilon=epsilon, V=Vint1, b=np.repeat(0, B.T.shape[1]).reshape(B.T.shape[1], 1), sigma_e2=0.1,
                          Lambda=np.repeat(10, len(ct_select)))
            # rownames(res$V) = colnames(Xinput_norm)
            # colnames(res$V) = colnames(B)
            ResList[str(iphi)] = res
            Obj = np.append(Obj, res["Obj"])
        self.Obj_hist = Obj
        Optimal_ix = np.where(Obj == Obj.max())[0][-1]  # in case if there are two equal objective function values
        OptimalPhi = phi[Optimal_ix]
        OptimalRes = ResList[str(Optimal_ix)]
        print("## Deconvolution Finish! ...\n")

        self.info_parameters["phi"] = OptimalPhi
        self.algorithm_matrix = {
            "B": (B.values * mean_B) / 1e-01,
            "Xinput_norm": (Xinput_norm.values * mean_X) / 1e-01,
            "Res": OptimalRes
        }

        self.spatial_location = spatial_location
        return self

    def predict(self):
        """Prediction function.

        Returns
        -------
        numpy.ndarray
            Predictions of cell-type proportions.

        """
        optim_res = self.algorithm_matrix["Res"]
        prop_pred = optim_res["V"] / optim_res["V"].sum(axis=1, keepdims=True)
        return prop_pred

    @staticmethod
    def score(x, y):
        """Model performance score measured by root mean square error.

        Parameters
        ----------
        x : np.ndarray
            Predicted cell-type proportion matrix (# spots x # cell-types)
        y : np.ndarray
            True cell-type proportion matrix (# spots x # cell-types)

        Returns
        -------
        float
            Root mean square error.

        """
        rmse = np.sqrt(((x - y / y.sum(1, keepdims=True))**2).sum() / y.size)
        return rmse



from dance.modules.multi_modality.match_modality.scmogcn import ScMoGCNWrapper
from dance.datasets.multimodality import ModalityMatchingDataset
from dance.transforms.graph_construct import basic_feature_propagation
from dance.utils import set_seed
import torch
import numpy as np
import random
import argparse

def normalize(X):
    return (X-X.mean())/X.std()

if __name__ == '__main__':
    rndseed = random.randint(0, 2147483647)
    parser = argparse.ArgumentParser()
    parser.add_argument('-t', '--subtask', default = 'openproblems_bmmc_cite_phase2_rna', choices = ['openproblems_bmmc_cite_phase2_rna', 'openproblems_bmmc_multiome_phase2_rna'])
    parser.add_argument('-d', '--data_folder', default = './data/modality_matching')
    parser.add_argument('-csv', '--csv_path', default = 'decoupled_lsi.csv')
    parser.add_argument('-l', '--layers', default = 4, type=int, choices = [3,4,5,6,7])
    parser.add_argument('-lr', '--learning_rate', default = 6e-4, type=float)
    parser.add_argument('-dis', '--disable_propagation', default = 0, type=int, choices = [0, 1, 2])
    parser.add_argument('-aux', '--auxiliary_loss', default = True, type=bool)
    parser.add_argument('-pk', '--pickle_suffix', default = '_lsi_input_pca_count.pkl')
    parser.add_argument('-seed', '--rnd_seed', default = rndseed, type=int)
    parser.add_argument('-cpu', '--cpus', default = 1, type = int)
    parser.add_argument('-device', '--device', default = 'cuda:3')
    parser.add_argument('-e', '--epochs', default = 2000, type = int)

    args = parser.parse_args()

    torch.set_num_threads(args.cpus)
    rndseed = args.rnd_seed
    set_seed(rndseed)

    subtask = args.subtask
    data_folder = args.data_folder
    device = args.device
    layers = args.layers
    pkl_path = subtask+args.pickle_suffix


    # DataLoader
    dataset = ModalityMatchingDataset(subtask, data_dir=data_folder).load_data().load_sol().preprocess('pca', pkl_path)

    if subtask == 'openproblems_bmmc_cite_phase2_rna':
        HIDDEN_SIZE = 64
        TEMPERATURE = 2.739896
        model = ScMoGCNWrapper(args, [[(dataset.preprocessed_features['mod1_train'].shape[1], 512, 0.25), (512, 512, 0.25), (512, HIDDEN_SIZE)],
                                      [(dataset.preprocessed_features['mod2_train'].shape[1], 512, 0.2), (512, 512, 0.2), (512, HIDDEN_SIZE)],
                                      [(HIDDEN_SIZE, 512, 0.2), (512, dataset.preprocessed_features['mod1_train'].shape[1])],
                                      [(HIDDEN_SIZE, 512, 0.2), (512, dataset.preprocessed_features['mod2_train'].shape[1])]], TEMPERATURE)
    else:
        HIDDEN_SIZE = 256
        TEMPERATURE = 3.065016
        model = ScMoGCNWrapper(args, [[(dataset.preprocessed_features['mod1_train'].shape[1], 1024, 0.5), (1024, 1024, 0.5), (1024, HIDDEN_SIZE)],
                                      [(dataset.preprocessed_features['mod2_train'].shape[1], 2048, 0.5), (2048, HIDDEN_SIZE)],
                                      [(HIDDEN_SIZE, 512, 0.2), (512, dataset.preprocessed_features['mod1_train'].shape[1])],
                                      [(HIDDEN_SIZE, 512, 0.2), (512, dataset.preprocessed_features['mod2_train'].shape[1])]], TEMPERATURE)

    hcell_mod1, hcell_mod2 = basic_feature_propagation(dataset, layers, device = device)
    z_test = torch.from_numpy(dataset.test_sol.X.toarray())
    labels1 = torch.argmax(z_test, dim=1).to(device)
    labels0 = torch.argmax(z_test, dim=0).to(device)

    model.fit(dataset, [hcell_mod1, hcell_mod2], [labels0, labels1])
    model.load(f'models/model_{rndseed}.pth')

    test_inputs = [hcell_mod1, hcell_mod2]
    test_idx = np.arange(dataset.sparse_features()[0].shape[0],
                    dataset.sparse_features()[0].shape[0] + dataset.sparse_features()[2].shape[0])
    print(model.predict(test_inputs, test_idx, enhance =True, dataset = dataset))
    print(model.score(test_inputs, test_idx,
                      z_test, enhance =True, dataset = dataset))




import warnings
warnings.filterwarnings("ignore")
from time import time
import os
import numpy as np
import scanpy as sc
import random
import torch

from dance.transforms.preprocess import normalize_adata, geneSelection
from dance.modules.single_modality.clustering.scdeepcluster import ScDeepCluster
from dance.datasets.singlemodality import ClusteringDataset

# for repeatability
random.seed(42)
np.random.seed(42)
torch.manual_seed(42)
torch.cuda.manual_seed(42)

if __name__ == "__main__":

    # setting the hyper parameters
    import argparse

    parser = argparse.ArgumentParser(description='train',
                                     formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument('--knn', default=20, type=int,
                        help='number of nearest neighbors, used by the Louvain algorithm')
    parser.add_argument('--resolution', default=.8, type=float,
                        help='resolution parameter, used by the Louvain algorithm, larger value for more number of clusters')
    parser.add_argument('--select_genes', default=0, type=int,
                        help='number of selected genes, 0 means using all genes')
    parser.add_argument('--batch_size', default=256, type=int)
    parser.add_argument('--data_dir', default='./data')
    parser.add_argument('--data_file', default='mouse_bladder_cell',
                        type=str)  # choice=['10X_PBMC', 'mouse_bladder_cell', 'mouse_ES_cell', 'worm_neuron_cell']
    parser.add_argument('--maxiter', default=500, type=int)
    parser.add_argument('--pretrain_epochs', default=50, type=int)
    parser.add_argument('--lr', default=0.1, type=float)
    parser.add_argument('--gamma', default=1., type=float,
                        help='coefficient of clustering loss')
    parser.add_argument('--sigma', default=2.5, type=float,
                        help='coefficient of random noise')
    parser.add_argument('--update_interval', default=1, type=int)
    parser.add_argument('--tol', default=0.001, type=float,
                        help='tolerance for delta clustering labels to terminate training stage')
    parser.add_argument('--ae_weights', default=None,
                        help='file to pretrained weights, None for a new pretraining')
    parser.add_argument('--save_dir', default='results/scDeepCluster/',
                        help='directory to save model weights during the training stage')
    parser.add_argument('--ae_weight_file', default='AE_weights.pth.tar',
                        help='file name to save model weights after the pretraining stage')
    parser.add_argument('--device', default='cuda')

    args = parser.parse_args()

    data = ClusteringDataset(args.data_dir, args.data_file).load_data()
    x = data.X
    y = data.Y
    n_clusters = len(np.unique(y))

    if args.select_genes > 0:
        importantGenes = geneSelection(x, n=args.select_genes, plot=False)
        x = x[:, importantGenes]

    # preprocessing scRNA-seq read counts matrix
    adata = sc.AnnData(x, dtype=np.float32)
    adata.obs['Group'] = y
    adata = adata.copy()
    adata.obs['DCA_split'] = 'train'
    adata.obs['DCA_split'] = adata.obs['DCA_split'].astype('category')

    adata = normalize_adata(adata,
                            size_factors=True,
                            normalize_input=True,
                            logtrans_input=True)

    input_size = adata.n_vars

    model = ScDeepCluster(input_dim=adata.n_vars, z_dim=32,
                          encodeLayer=[256, 64], decodeLayer=[64, 256],
                          sigma=args.sigma, gamma=args.gamma, device=args.device)

    t0 = time()
    if args.ae_weights is None:
        model.pretrain_autoencoder(X=adata.X, X_raw=adata.raw.X, size_factor=adata.obs.size_factors,
                                   batch_size=args.batch_size, epochs=args.pretrain_epochs,
                                   ae_weights=args.ae_weight_file)
    else:
        if os.path.isfile(args.ae_weights):
            print("==> loading checkpoint '{}'".format(args.ae_weights))
            checkpoint = torch.load(args.ae_weights)
            model.load_state_dict(checkpoint['ae_state_dict'])
        else:
            print("==> no checkpoint found at '{}'".format(args.ae_weights))
            raise ValueError

    print('Pretraining time: %d seconds.' % int(time() - t0))

    if not os.path.exists(args.save_dir):
        os.makedirs(args.save_dir)

    model.fit(X=adata.X, X_raw=adata.raw.X, size_factor=adata.obs.size_factors,
              n_clusters=n_clusters, init_centroid=None, y_pred_init=None, y=y,
              lr=args.lr, batch_size=args.batch_size, num_epochs=args.maxiter,
              update_interval=args.update_interval, tol=args.tol, save_dir=args.save_dir)

    print('Total time: %d seconds.' % int(time() - t0))

    y_pred = model.predict()
    #    print(f'Prediction: {y_pred}')

    acc, nmi, ari = model.score(y)
    print("ACC: %.4f, NMI: %.4f, ARI: %.4f" % (acc, nmi, ari))

""" Reproduction information
10X PBMC:
python scdeepcluster.py --data_file='10X_PBMC'

Mouse ES:
python scdeepcluster.py --data_file='mouse_ES_cell'

Worm Neuron:
python scdeepcluster.py --data_file='worm_neuron_cell'

Mouse Bladder:
python scdeepcluster.py --data_file='mouse_bladder_cell' --sigma=2.75
"""


import torch.nn as nn
import torch.nn.functional as F
import torch
import math
import torch.optim as optim
import numpy as np
import os
from torch.utils.data import DataLoader
from dance.utils.metrics import *
from dance.utils import SimpleIndexDataset

def propagation_layer_combination(X, idx, wt, from_logits=True):
    if from_logits:
        wt = torch.softmax(wt, -1)

    x = 0
    for i in range(wt.shape[0]):
        x += wt[i] * X[i][idx]

    return x

class ScMoGCNWrapper(object):
    """ScMoGCN class.

    Parameters
    ----------
    args : argparse.Namespace
        A Namespace object that contains arguments of ScMoGCN. For details of parameters in parser args, please refer to link (parser help document).
    dataset : dance.datasets.multimodality.JointEmbeddingNIPSDataset
        Joint embedding dataset.
    """
    def __init__(self, args, dataset):
        self.model = Transformation(dataset.nb_cell_types, dataset.nb_batches, dataset.nb_phases,
                                    dataset.preprocessed_data['X_train'].shape[1]).to(args.device)
        self.args = args
        self.wt = torch.tensor([0.] * (args.layers - 1)).to(args.device).requires_grad_(True)


    def fit(self, dataset, inputs, labels):
        """fit function for training.

        Parameters
        ----------
        dataset : dance.datasets.multimodality.JointEmbeddingNIPSDataset
            Modality features.
        inputs : torch.Tensor
            Modality features.
        labels : list[torch.Tensor]
            Multiple auxiliary labels for supervision.

        Returns
        -------
        None.
        """

        wt = self.wt
        X = inputs
        Y = [torch.from_numpy(labels[0]).long().to(self.args.device),
                             torch.from_numpy(labels[1]).long().to(self.args.device),
                             torch.from_numpy(labels[3]).float().to(self.args.device)]

        idx = np.random.permutation(dataset.preprocessed_data['X_train'].shape[0])
        train_idx = idx[:int(idx.shape[0] * 0.9)]
        val_idx = idx[int(idx.shape[0] * 0.9):]

        train_dataset = SimpleIndexDataset(train_idx)
        train_loader = DataLoader(dataset=train_dataset,
                                  batch_size=self.args.batch_size,
                                  shuffle=True,
                                  num_workers=1, )

        ce = nn.CrossEntropyLoss()
        mse = nn.MSELoss()

        optimizer = torch.optim.AdamW([{'params': self.model.parameters()}, {'params': wt, 'weight_decay': 0}], lr=1e-4)
        vals = []

        for epoch in range(60):
            self.model.train()
            total_loss = [0] * 5
            print('epoch', epoch)

            for iter, batch_idx in enumerate(train_loader):


                batch_x = propagation_layer_combination(X, batch_idx, wt)
                batch_y = [batch_x, Y[0][batch_idx], Y[1][batch_idx],
                           Y[2][batch_idx]]

                output = self.model(batch_x)

                # loss1 = mse(output[0], batch_y[0]) # option 1: recover features after propagation
                loss1 = mse(output[0], batch_x)  # option 2: recover Isi features
                loss2 = ce(output[1], batch_y[1])
                loss3 = torch.norm(output[2], p=2, dim=-1).sum() * 1e-2
                loss4 = mse(output[3], batch_y[3])

                loss = loss1 * 0.7 + loss2 * 0.2 + loss3 * 0.05 + loss4 * 0.05

                total_loss[0] += loss1.item()
                total_loss[1] += loss2.item()
                total_loss[2] += loss3.item()
                total_loss[3] += loss4.item()
                total_loss[4] += loss.item()

                optimizer.zero_grad()
                loss.backward()
                optimizer.step()

            for i in range(4):
                print(f'loss{i + 1}', total_loss[i] / len(train_loader), end=', ')
            print()

            loss1, loss2, loss3, loss4 = self.score(X, val_idx, Y)
            weighted_loss = loss1 * 0.7 + loss2 * 0.2 + loss3 * 0.05 + loss4 * 0.05
            print('val-loss1', loss1, 'val-loss2', loss2, 'val-loss3', loss3, 'val-loss4', loss4)
            print('val score', weighted_loss)
            vals.append(weighted_loss)
            if min(vals) == vals[-1]:
                if not os.path.exists('models'):
                    os.mkdir('models')
                torch.save(self.model.state_dict(), f'models/model_joint_embedding_{self.args.rnd_seed}.pth')
                weight_record = wt.detach()

            if min(vals) != min(vals[-10:]):
                break


        self.wt = weight_record

    def to(self, device):
        """Performs device conversion.

        Parameters
        ----------
        device : str
            Target device.

        Returns
        -------
        self : ScMoGCNWrapper
            Converted model.
        """

        self.args.device = device
        self.model = self.model.to(device)
        return self

    def load(self, path, map_location = None):
        """Load model parameters from checkpoint file.

        Parameters
        ----------
        path : str
            Path to the checkpoint file.
        map_location : str optional
            Mapped device. This parameter will be passed to torch.load function if not none.

        Returns
        -------
        None.

        """
        if map_location is not None:
            self.model.load_state_dict(torch.load(path, map_location=map_location))
        else:
            self.model.load_state_dict(torch.load(path))

    def predict(self, inputs, idx):
        """Predict function to get latent representation of data.

        Parameters
        ----------
        inputs : torch.Tensor
            Multimodality features.

        Returns
        -------
        prediction : torch.Tensor
            Joint embedding of input data.

        """
        self.model.eval()
        wt = self.wt

        with torch.no_grad():
            X = propagation_layer_combination(inputs, idx, wt)

        return self.model.encoder(X)


    def score(self, inputs, idx, labels, metric = 'loss'):
        """Score function to get score of prediction.

        Parameters
        ----------
        inputs : torch.Tensor
            Multimodality features.
        idx : Iterable[int]
            Index of testing samples for scoring.
        labels: list[torch.Tensor]
            Multiple cell labels for evaluation.
        metric : str optional
            The type of evaluation metric, by default to be 'loss'.

        Returns
        -------
        loss1 : float
            Reconstruction loss.
        loss2 : float
            Cell type classfication loss.
        loss3 : float
            Batch regularization loss.
        loss4 : float
            Cell cycle score loss.

        """

        self.model.eval()
        ce = nn.CrossEntropyLoss()
        mse = nn.MSELoss()

        with torch.no_grad():
            if metric=='loss':
                X = propagation_layer_combination(inputs, idx, self.wt)
                output = self.model(X)
                loss1 = mse(output[0], X).item()
                loss2 = ce(output[1], labels[0][idx]).item()
                loss3 = (torch.norm(output[2], p=2, dim=-1).sum() * 1e-2).item()
                loss4 = mse(output[3], labels[2][idx]).item()

                return loss1, loss2, loss3, loss4
            else:
                emb = self.predict(inputs, idx).cpu().numpy()
                kmeans = KMeans(n_clusters=10, n_init=5, random_state=200)

                # adata.obs['batch'] = adata_sol.obs['batch'][adata.obs_names]
                # adata.obs['cell_type'] = adata_sol.obs['cell_type'][adata.obs_names]
                true_labels = labels
                pred_labels = kmeans.fit_predict(emb)
                NMI_score = round(normalized_mutual_info_score(true_labels, pred_labels, average_method='max'), 3)
                ARI_score = round(adjusted_rand_score(true_labels, pred_labels), 3)

                # print('ARI: ' + str(ARI_score) + ' NMI: ' + str(NMI_score))
                return NMI_score, ARI_score

class Transformation(nn.Module):
    def __init__(self, nb_cell_types, nb_batches, nb_phases, input_dimension):
        super().__init__()
        self.nb_cell_types = nb_cell_types
        self.nb_batches = nb_batches
        self.nb_phases = nb_phases

        self.linear1 = nn.Linear(input_dimension, 150)
        self.linear2 = nn.Linear(150, 120)
        self.linear3 = nn.Linear(120, 100)
        self.linear4 = nn.Linear(100, 61)

        self.bn1 = nn.BatchNorm1d(150)
        self.bn2 = nn.BatchNorm1d(120)
        self.bn3 = nn.BatchNorm1d(100)

        self.act1 = nn.GELU()
        self.act2 = nn.GELU()
        self.act3 = nn.GELU()

        self.decoder = nn.Sequential(nn.Linear(61, 150),
                                     nn.ReLU(),
                                     nn.Linear(150, input_dimension),
                                     nn.ReLU(), )

    def encoder(self, x):
        x = self.linear1(x)
        x = self.act1(x)
        x = self.bn1(x)
        x = F.dropout(x, p=0.3, training=self.training)

        x = self.linear2(x)
        x = self.act2(x)
        x = self.bn2(x)
        x = F.dropout(x, p=0.3, training=self.training)

        x = self.linear3(x)
        x = self.act3(x)
        x = self.bn3(x)
        x = F.dropout(x, p=0.3, training=self.training)

        x = self.linear4(x)
        return x

    def forward(self, x):
        x = self.encoder(x)
        x0 = x
        x = self.decoder(x)

        return x, x0[:, :self.nb_cell_types], x0[:, self.nb_cell_types:self.nb_cell_types + self.nb_batches], \
               x0[:, self.nb_cell_types + self.nb_batches:self.nb_cell_types + self.nb_batches + self.nb_phases]




import argparse
import random
from pprint import pprint

import numpy as np
import torch

from dance.datasets.singlemodality import CellTypeDataset
from dance.modules.single_modality.cell_type_annotation.scdeepsort import ScDeepSort

if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("--random_seed", type=int, default=10)
    parser.add_argument("--data_type", type=str, default="scdeepsort_exp")
    parser.add_argument("--dropout", type=float, default=0.1, help="dropout probability")
    parser.add_argument("--gpu", type=int, default=-1, help="GPU id, -1 for cpu")
    parser.add_argument("--filetype", default="csv", type=str, choices=["csv", "gz"], help="data file type, csv or gz")
    parser.add_argument("--lr", type=float, default=1e-3, help="learning rate")
    parser.add_argument("--weight_decay", type=float, default=5e-4, help="Weight for L2 loss")
    parser.add_argument("--n_epochs", type=int, default=300, help="number of training epochs")
    parser.add_argument("--dense_dim", type=int, default=400, help="number of hidden gcn units")
    parser.add_argument("--hidden_dim", type=int, default=200, help="number of hidden gcn units")
    parser.add_argument("--n_layers", type=int, default=1, help="number of hidden gcn layers")
    parser.add_argument("--threshold", type=float, default=0,
                        help="the threshold to connect edges between cells and genes")
    parser.add_argument("--num_neighbors", type=int, default=0,
                        help="number of neighbors to sample in message passing process. 0 means all neighbors")
    parser.add_argument("--exclude_rate", type=float, default=0.005, help="exclude some cells less than this rate.")
    parser.add_argument("--species", default="mouse", type=str)
    parser.add_argument("--tissue", default="Spleen", type=str)
    parser.add_argument("--batch_size", type=int, default=500)
    parser.add_argument("--unsure_rate", type=float, default=2.0, help="the threshold to predict unsure cell")
    parser.add_argument("--evaluate", dest="evaluate", action="store_true")
    parser.add_argument("--test", dest="evaluate", action="store_false")
    parser.set_defaults(evaluate=True)
    parser.add_argument("--test_dir", default="test", type=str, help="test directory")
    parser.add_argument("--save_dir", default="result", type=str, help="save directory")
    parser.add_argument("--test_rate", type=float, default=0.2)
    parser.add_argument("--test_dataset", nargs="+", type=int, default=[1759], help="Testing dataset IDs")
    params = parser.parse_args()
    pprint(vars(params))

    dataloader = CellTypeDataset(data_type="scdeepsort_exp", random_seed=params.random_seed, dense_dim=params.dense_dim,
                                 test_dataset=params.test_dataset, species=params.species, tissue=params.tissue,
                                 gpu=params.gpu, evaluate=params.evaluate, test_dir=params.test_dir,
                                 filetype=params.filetype, threshold=params.threshold, exclude_rate=params.exclude_rate,
                                 test_rate=params.test_rate, score=True)
    dataloader = dataloader.load_data()
    random.seed(params.random_seed)
    np.random.seed(params.random_seed)
    torch.manual_seed(params.random_seed)
    torch.cuda.manual_seed(params.random_seed)
    trainer = ScDeepSort(params)

    trainer.fit(dataloader.num_cells, dataloader.num_genes, dataloader.num_labels, dataloader.graph,
                dataloader.train_ids, dataloader.test_ids, dataloader.labels)
    prediction_labels = trainer.predict(dataloader.id2label_test, dataloader.test_dict)
    evaluation_scores = trainer.score(prediction_labels, dataloader.test_label_dict)
    pprint(evaluation_scores)

"""To reproduce the benchmarking results, please run the following commands:

python scdeepsort.py --data_type scdeepsort --tissue Brain --test_data 2695

python scdeepsort.py --data_type scdeepsort --tissue Spleen --test_data 1759

python scdeepsort.py --data_type scdeepsort --tissue Kidney --test_data 203

"""



import numpy as np
import pandas as pd
import os
import time
import math
import torch.utils.data as data_utils
import collections

from torch.autograd import Variable
from torch import optim
from sklearn.cluster import KMeans
from sklearn import metrics
from sklearn.metrics import cohen_kappa_score
from sklearn.metrics.cluster import normalized_mutual_info_score
from sklearn.mixture import GaussianMixture
from collections import OrderedDict

import torch
from torch import nn
from torch.nn import functional as F
from torch.distributions import Normal, kl_divergence as kl

# from DCCA.loss_function import log_zinb_positive, log_nb_positive, binary_cross_entropy, mse_loss, KL_diver
from dance.utils.loss import NSTLoss, FactorTransfer, Similarity, Correlation, Attention, Eucli_dis, L1_dis, KL_diver
import warnings

warnings.filterwarnings("ignore", category=DeprecationWarning)

def mse_loss(y_true, y_pred):

    mask = torch.sign(y_true)

    y_pred = y_pred.float()
    y_true = y_true.float()

    ret = torch.pow( (y_pred - y_true) * mask , 2)

    return torch.sum( ret, dim = 1 )

def binary_cross_entropy(recon_x, x):
    # mask = torch.sign(x)
    return - torch.sum(x * torch.log(recon_x + 1e-8) + (1 - x) * torch.log(1 - recon_x + 1e-8), dim=1)


def log_zinb_positive(x, mu, theta, pi, eps=1e-8):
    x = x.float()

    if theta.ndimension() == 1:
        theta = theta.view(1, theta.size(0))

    softplus_pi = F.softplus(-pi)

    log_theta_eps = torch.log(theta + eps)

    log_theta_mu_eps = torch.log(theta + mu + eps)

    pi_theta_log = -pi + theta * (log_theta_eps - log_theta_mu_eps)

    case_zero = F.softplus(pi_theta_log) - softplus_pi
    mul_case_zero = torch.mul((x < eps).type(torch.float32), case_zero)

    case_non_zero = (-softplus_pi + pi_theta_log
                     + x * (torch.log(mu + eps) - log_theta_mu_eps)
                     + torch.lgamma(x + theta)
                     - torch.lgamma(theta)
                     - torch.lgamma(x + 1))

    mul_case_non_zero = torch.mul((x > eps).type(torch.float32), case_non_zero)

    res = mul_case_zero + mul_case_non_zero

    return - torch.sum(res, dim=1)


def log_nb_positive(x, mu, theta, eps=1e-8):
    x = x.float()

    if theta.ndimension() == 1:
        theta = theta.view(
            1, theta.size(0)
        )  # In this case, we reshape theta for broadcasting

    log_theta_mu_eps = torch.log(theta + mu + eps)

    res = (
            theta * (torch.log(theta + eps) - log_theta_mu_eps)
            + x * (torch.log(mu + eps) - log_theta_mu_eps)
            + torch.lgamma(x + theta)
            - torch.lgamma(theta)
            - torch.lgamma(x + 1)
    )

    return - torch.sum(res, dim=1)


def build_multi_layers(layers, use_batch_norm=True, dropout_rate=0.1):
    """Build multilayer linear perceptron"""

    if dropout_rate > 0:
        fc_layers = nn.Sequential(
            collections.OrderedDict(
                [
                    (
                        "Layer {}".format(i),
                        nn.Sequential(
                            nn.Linear(n_in, n_out),
                            nn.BatchNorm1d(n_out, momentum=0.01, eps=0.001),
                            nn.ReLU(),
                            nn.Dropout(p=dropout_rate),
                        ),
                    )

                    for i, (n_in, n_out) in enumerate(zip(layers[:-1], layers[1:]))
                ]
            )
        )

    else:
        fc_layers = nn.Sequential(
            collections.OrderedDict(
                [
                    (
                        "Layer {}".format(i),
                        nn.Sequential(
                            nn.Linear(n_in, n_out),
                            nn.BatchNorm1d(n_out, momentum=0.01, eps=0.001),
                            nn.ReLU(),
                        ),
                    )

                    for i, (n_in, n_out) in enumerate(zip(layers[:-1], layers[1:]))
                ]
            )
        )

    return fc_layers

def adjust_learning_rate(init_lr, optimizer, iteration, max_lr, adjust_epoch):

    lr = max(init_lr * (0.9 ** (iteration//adjust_epoch)), max_lr)
    for param_group in optimizer.param_groups:
        param_group["lr"] = lr

    return lr

class Encoder(nn.Module):

    ## for one modulity
    def __init__(self, layer, hidden, Z_DIMS, droprate=0.1):
        super(Encoder, self).__init__()

        if len(layer) > 1:
            self.fc1 = build_multi_layers(layers=layer, dropout_rate=droprate)

        self.layer = layer
        self.fc_means = nn.Linear(hidden, Z_DIMS)
        self.fc_logvar = nn.Linear(hidden, Z_DIMS)

    def reparametrize(self, means, logvar):

        if self.training:
            std = logvar.mul(0.5).exp_()
            eps = Variable(std.data.new(std.size()).normal_())
            return eps.mul(std).add_(means)
        else:
            return means

    def return_all_params(self, x):

        if len(self.layer) > 1:
            h = self.fc1(x)
        else:
            h = x

        mean_x = self.fc_means(h)
        logvar_x = self.fc_logvar(h)
        latent = self.reparametrize(mean_x, logvar_x)

        return mean_x, logvar_x, latent, h

    def forward(self, x):

        _, _, latent = self.return_all_params(x)

        return latent


class DecoderLogNormZINB(nn.Module):

    ### for scRNA-seq, refered by DCA

    def __init__(self, layer, hidden, input_size, droprate=0.1):
        super(DecoderLogNormZINB, self).__init__()

        self.decoder = build_multi_layers(layers=layer, dropout_rate=droprate)

        self.decoder_scale = nn.Linear(hidden, input_size)
        self.decoder_r = nn.Linear(hidden, input_size)
        self.dropout = nn.Linear(hidden, input_size)

    def forward(self, z=None, scale_factor=1.0):
        latent = self.decoder(z)

        normalized_x = F.softmax(self.decoder_scale(latent), dim=1)

        batch_size = normalized_x.size(0)
        scale_factor.resize_(batch_size, 1)
        scale_factor.repeat(1, normalized_x.size(1))

        scale_x = torch.exp(scale_factor) * normalized_x  ###

        disper_x = torch.exp(self.decoder_r(latent))  ### theta
        dropout_rate = self.dropout(latent)

        return dict(normalized=normalized_x,
                    disperation=disper_x,
                    dropoutrate=dropout_rate,
                    scale_x=scale_x
                    )


class DecoderLogNormNB(nn.Module):

    ### for scRNA-seq

    def __init__(self, layer, hidden, input_size, droprate=0.1):
        super(DecoderLogNormNB, self).__init__()

        self.decoder = build_multi_layers(layers=layer, dropout_rate=droprate)

        self.decoder_scale = nn.Linear(hidden, input_size)
        self.decoder_r = nn.Linear(hidden, input_size)

    def forward(self, z, scale_factor=torch.tensor(1.0)):
        latent = self.decoder(z)

        normalized_x = F.softmax(self.decoder_scale(latent), dim=1)  ## mean gamma

        batch_size = normalized_x.size(0)
        scale_factor.resize_(batch_size, 1)
        scale_factor.repeat(1, normalized_x.size(1))

        scale_x = torch.exp(scale_factor) * normalized_x

        disper_x = torch.exp(self.decoder_r(latent))  ### theta

        return dict(normalized=normalized_x,
                    disperation=disper_x,
                    scale_x=scale_x,
                    )


class Decoder(nn.Module):
    ### for scATAC-seq
    def __init__(self, layer, hidden, input_size, Type="Bernoulli", droprate=0.1):
        super(Decoder, self).__init__()

        if len(layer) > 1:
            self.decoder = build_multi_layers(layer, dropout_rate=droprate)

        self.decoder_x = nn.Linear(hidden, input_size)
        self.Type = Type
        self.layer = layer

    def forward(self, z):

        if len(self.layer) > 1:
            latent = self.decoder(z)
        else:
            latent = z

        recon_x = self.decoder_x(latent)

        if self.Type == "Bernoulli":
            Final_x = torch.sigmoid(recon_x)

        elif self.Type == "Gaussian1":
            Final_x = F.softmax(recon_x, dim=1)

        elif self.Type == "Gaussian":
            Final_x = torch.sigmoid(recon_x)

        elif self.Type == "Gaussian2":
            Final_x = F.relu(recon_x)

        else:
            Final_x = recon_x

        return Final_x

class VAE(nn.Module):
    # def __init__( self, layer_e, hidden1, hidden2, layer_l, layer_d, hidden ):
    def __init__(self, layer_e, hidden1, Zdim, layer_d, hidden2, Type='NB', droprate=0.1):

        super(VAE, self).__init__()

        ###  encoder
        self.encoder = Encoder(layer_e, hidden1, Zdim, droprate=droprate)
        self.activation = nn.Softmax(dim=-1)

        ### the decoder
        if Type == 'ZINB':
            self.decoder = DecoderLogNormZINB(layer_d, hidden2, layer_e[0], droprate=droprate)

        elif Type == 'NB':
            self.decoder = DecoderLogNormNB(layer_d, hidden2, layer_e[0], droprate=droprate)

        else:  ## Bernoulli, or Gaussian
            self.decoder = Decoder(layer_d, hidden2, layer_e[0], Type, droprate=droprate)

        ### parameters
        self.Type = Type

    def inference(self, X=None, scale_factor=1.0):
        # encoder
        mean_1, logvar_1, latent_1, hidden = self.encoder.return_all_params(X)

        ### decoder
        if self.Type == 'ZINB':
            output = self.decoder(latent_1, scale_factor)
            norm_x = output["normalized"]
            disper_x = output["disperation"]
            recon_x = output["scale_x"]
            dropout_rate = output["dropoutrate"]

        elif self.Type == 'NB':
            output = self.decoder(latent_1, scale_factor)
            norm_x = output["normalized"]
            disper_x = output["disperation"]
            recon_x = output["scale_x"]
            dropout_rate = None

        else:
            recons_x = self.decoder(latent_1)
            recon_x = recons_x
            norm_x = recons_x
            disper_x = None
            dropout_rate = None

        return dict(norm_x=norm_x, disper_x=disper_x, dropout_rate=dropout_rate,
                    recon_x=recon_x, latent_z1=latent_1, mean_1=mean_1,
                    logvar_1=logvar_1, hidden=hidden
                    )

    def return_loss(self, X=None, X_raw=None, latent_pre=None,
                    mean_pre=None, logvar_pre=None, latent_pre_hidden=None,
                    scale_factor=1.0, cretion_loss=None, attention_loss=None):

        output = self.inference(X, scale_factor)
        recon_x = output["recon_x"]
        disper_x = output["disper_x"]
        dropout_rate = output["dropout_rate"]

        mean_1 = output["mean_1"]
        logvar_1 = output["logvar_1"]
        latent_z1 = output["latent_z1"]

        hidden = output["hidden"]

        if self.Type == 'ZINB':
            loss = log_zinb_positive(X_raw, recon_x, disper_x, dropout_rate)

        elif self.Type == 'NB':
            loss = log_nb_positive(X_raw, recon_x, disper_x)

        elif self.Type == 'Bernoulli':  # here X and X_raw are same
            loss = binary_cross_entropy(recon_x, X_raw)

        else:
            loss = mse_loss(X, recon_x)

        ##calculate KL loss for Gaussian distribution
        mean = torch.zeros_like(mean_1)
        scale = torch.ones_like(logvar_1)
        kl_divergence_z = kl(Normal(mean_1, torch.exp(logvar_1)),
                             Normal(mean, scale)).sum(dim=1)

        atten_loss1 = torch.tensor(0.0)
        if latent_pre is not None and latent_pre_hidden is not None:

            if attention_loss == "KL_div":
                atten_loss1 = cretion_loss(mean_1, logvar_1, mean_pre, logvar_pre)

            else:
                atten_loss1 = cretion_loss(latent_z1, latent_pre)

        return loss, kl_divergence_z, atten_loss1

    def forward(self, X=None, scale_factor=1.0):

        output = self.inference(X, scale_factor)

        return output

    def fit(self, train_loader, test_loader, total_loader, model_pre, args, criterion, cycle, state, first="RNA",
            attention_loss="Eucli"):

        params = filter(lambda p: p.requires_grad, self.parameters())

        if cycle % 2 == 0:
            optimizer = optim.Adam(params, lr=args.lr1, weight_decay=args.weight_decay, eps=args.eps)
        else:
            optimizer = optim.Adam(params, lr=args.lr2, weight_decay=args.weight_decay, eps=args.eps)

        train_loss_list = []
        reco_epoch_test = 0
        test_like_max = 100000
        flag_break = 0

        patience_epoch = 0
        args.anneal_epoch = 10

        model_pre.eval()

        start = time.time()

        for epoch in range(1, args.max_epoch + 1):

            self.train()

            patience_epoch += 1
            kl_weight = min(1, epoch / args.anneal_epoch)

            if cycle % 2 == 0:
                epoch_lr = adjust_learning_rate(args.lr1, optimizer, epoch, args.flr1, 10)
            else:
                epoch_lr = adjust_learning_rate(args.lr2, optimizer, epoch, args.flr2, 10)

            for batch_idx, (X1, X1_raw, size_factor1, X2, X2_raw, size_factor2) in enumerate(train_loader):

                X1, X1_raw, size_factor1 = X1.to(args.device), X1_raw.to(args.device), size_factor1.to(args.device)
                X2, X2_raw, size_factor2 = X2.to(args.device), X2_raw.to(args.device), size_factor2.to(args.device)

                X1, X1_raw, size_factor1 = Variable(X1), Variable(X1_raw), Variable(size_factor1)
                X2, X2_raw, size_factor2 = Variable(X2), Variable(X2_raw), Variable(size_factor2)

                optimizer.zero_grad()

                if first == "RNA":

                    if cycle % 2 == 0:

                        if state == 0:
                            # initialization of scRNA-seq model
                            loss1, kl_divergence_z, atten_loss1 = self.return_loss(X1, X1_raw, None, None, None, None,
                                                                                   size_factor1, criterion,
                                                                                   attention_loss)
                            loss = torch.mean(loss1 + (kl_weight * kl_divergence_z))

                        else:
                            # transfer representation from scEpigenomics model to scRNA-seq model
                            result_2 = model_pre(X2, size_factor2)
                            latent_z1 = result_2["latent_z1"].to(args.device)
                            hidden_1 = result_2["hidden"].to(args.device)
                            mean_1 = result_2["mean_1"].to(args.device)
                            logvar_1 = result_2["logvar_1"].to(args.device)

                            loss1, kl_divergence_z, atten_loss1 = self.return_loss(X1, X1_raw, latent_z1, mean_1,
                                                                                   logvar_1, hidden_1,
                                                                                   size_factor1, criterion,
                                                                                   attention_loss)
                            loss = torch.mean(loss1 + (kl_weight * kl_divergence_z) + (args.sf2 * (atten_loss1)))

                    else:
                        if state == 0:
                            # initialization of scEpigenomics model
                            loss1, kl_divergence_z, atten_loss1 = self.return_loss(X2, X2_raw, None, None, None, None,
                                                                                   size_factor2, criterion,
                                                                                   attention_loss)
                            loss = torch.mean(loss1 + (kl_weight * kl_divergence_z))

                        else:
                            # transfer representation form scRNA-seq model to scEpigenomics model
                            result_2 = model_pre(X1, size_factor1)
                            latent_z1 = result_2["latent_z1"].to(args.device)
                            hidden_1 = result_2["hidden"].to(args.device)
                            mean_1 = result_2["mean_1"].to(args.device)
                            logvar_1 = result_2["logvar_1"].to(args.device)

                            loss1, kl_divergence_z, atten_loss1 = self.return_loss(X2, X2_raw, latent_z1, mean_1,
                                                                                   logvar_1, hidden_1,
                                                                                   size_factor2, criterion,
                                                                                   attention_loss)
                            loss = torch.mean(loss1 + (kl_weight * kl_divergence_z) + (args.sf1 * (atten_loss1)))
                else:

                    if cycle % 2 == 0:

                        if state == 0:
                            # initialization of scEpigenomics model
                            loss1, kl_divergence_z, atten_loss1 = self.return_loss(X2, X2_raw, None, None, None, None,
                                                                                   size_factor2, criterion,
                                                                                   attention_loss)
                            loss = torch.mean(loss1 + (kl_weight * kl_divergence_z))

                        else:
                            # transfer representation from scRNA-seq model to scEpigenomics model
                            result_2 = model_pre(X1, size_factor1)
                            latent_z1 = result_2["latent_z1"].to(args.device)
                            hidden_1 = result_2["hidden"].to(args.device)
                            mean_1 = result_2["mean_1"].to(args.device)
                            logvar_1 = result_2["logvar_1"].to(args.device)

                            loss1, kl_divergence_z, atten_loss1 = self.return_loss(X2, X2_raw, latent_z1, mean_1,
                                                                                   logvar_1, hidden_1,
                                                                                   size_factor2, criterion,
                                                                                   attention_loss)
                            loss = torch.mean(loss1 + (kl_weight * kl_divergence_z) + (args.sf1 * (atten_loss1)))

                    else:
                        if state == 0:
                            # initialization of scRNA-seq model
                            loss1, kl_divergence_z, atten_loss1 = self.return_loss(X1, X1_raw, None, None, None, None,
                                                                                   size_factor1, criterion,
                                                                                   attention_loss)
                            loss = torch.mean(loss1 + (kl_weight * kl_divergence_z))

                        else:
                            # transfer representation from scEpigenomics model to scRNA-seq model
                            result_2 = model_pre(X2, size_factor2)
                            latent_z1 = result_2["latent_z1"].to(args.device)
                            hidden_1 = result_2["hidden"].to(args.device)
                            mean_1 = result_2["mean_1"].to(args.device)
                            logvar_1 = result_2["logvar_1"].to(args.device)

                            loss1, kl_divergence_z, atten_loss1 = self.return_loss(X1, X1_raw, latent_z1, mean_1,
                                                                                   logvar_1, hidden_1,
                                                                                   size_factor1, criterion,
                                                                                   attention_loss)
                            loss = torch.mean(loss1 + (kl_weight * kl_divergence_z) + (args.sf2 * (atten_loss1)))

                loss.backward()
                optimizer.step()

            if epoch % args.epoch_per_test == 0 and epoch > 0:
                self.eval()

                with torch.no_grad():

                    for batch_idx, (X1, X1_raw, size_factor1, X2, X2_raw, size_factor2) in enumerate(test_loader):

                        X1, X1_raw, size_factor1 = X1.to(args.device), X1_raw.to(args.device), size_factor1.to(args.device)
                        X2, X2_raw, size_factor2 = X2.to(args.device), X2_raw.to(args.device), size_factor2.to(args.device)

                        X1, X1_raw, size_factor1 = Variable(X1), Variable(X1_raw), Variable(size_factor1)
                        X2, X2_raw, size_factor2 = Variable(X2), Variable(X2_raw), Variable(size_factor2)

                        if first == "RNA":

                            if cycle % 2 == 0:
                                if state == 0:
                                    loss1, kl_divergence_z, atten_loss1 = self.return_loss(X1, X1_raw, None, None, None,
                                                                                           None,
                                                                                           size_factor1, criterion,
                                                                                           attention_loss)
                                    test_loss = torch.mean(loss1 + (kl_weight * kl_divergence_z))

                                else:
                                    result_2 = model_pre(X2, size_factor2)
                                    latent_z1 = result_2["latent_z1"].to(args.device)
                                    hidden_1 = result_2["hidden"].to(args.device)
                                    mean_1 = result_2["mean_1"].to(args.device)
                                    logvar_1 = result_2["logvar_1"].to(args.device)

                                    loss1, kl_divergence_z, atten_loss1 = self.return_loss(X1, X1_raw, latent_z1,
                                                                                           mean_1, logvar_1, hidden_1,
                                                                                           size_factor1, criterion,
                                                                                           attention_loss)
                                    test_loss = torch.mean(
                                        loss1 + (kl_weight * kl_divergence_z) + (args.sf2 * (atten_loss1)))

                            else:
                                if state == 0:
                                    loss1, kl_divergence_z, atten_loss1 = self.return_loss(X2, X2_raw, None, None, None,
                                                                                           None,
                                                                                           size_factor2, criterion,
                                                                                           attention_loss)
                                    test_loss = torch.mean(loss1 + (kl_weight * kl_divergence_z))

                                else:
                                    result_2 = model_pre(X1, size_factor1)
                                    latent_z1 = result_2["latent_z1"].to(args.device)
                                    hidden_1 = result_2["hidden"].to(args.device)
                                    mean_1 = result_2["mean_1"].to(args.device)
                                    logvar_1 = result_2["logvar_1"].to(args.device)

                                    loss1, kl_divergence_z, atten_loss1 = self.return_loss(X2, X2_raw, latent_z1,
                                                                                           mean_1, logvar_1, hidden_1,
                                                                                           size_factor2, criterion,
                                                                                           attention_loss)
                                    test_loss = torch.mean(
                                        loss1 + (kl_weight * kl_divergence_z) + (args.sf1 * (atten_loss1)))

                        else:
                            if cycle % 2 == 0:

                                if state == 0:
                                    loss1, kl_divergence_z, atten_loss1 = self.return_loss(X2, X2_raw, None, None, None,
                                                                                           None,
                                                                                           size_factor2, criterion,
                                                                                           attention_loss)
                                    test_loss = torch.mean(loss1 + (kl_weight * kl_divergence_z))

                                else:
                                    result_2 = model_pre(X1, size_factor1)
                                    latent_z1 = result_2["latent_z1"].to(args.device)
                                    hidden_1 = result_2["hidden"].to(args.device)
                                    mean_1 = result_2["mean_1"].to(args.device)
                                    logvar_1 = result_2["logvar_1"].to(args.device)

                                    loss1, kl_divergence_z, atten_loss1 = self.return_loss(X2, X2_raw, latent_z1,
                                                                                           mean_1, logvar_1, hidden_1,
                                                                                           size_factor2, criterion,
                                                                                           attention_loss)
                                    test_loss = torch.mean(
                                        loss1 + (kl_weight * kl_divergence_z) + (args.sf1 * (atten_loss1)))

                            else:
                                if state == 0:
                                    loss1, kl_divergence_z, atten_loss1 = self.return_loss(X1, X1_raw, None, None, None,
                                                                                           None,
                                                                                           size_factor1, criterion,
                                                                                           attention_loss)
                                    test_loss = torch.mean(loss1 + (kl_weight * kl_divergence_z))

                                else:
                                    result_2 = model_pre(X2, size_factor2)
                                    latent_z1 = result_2["latent_z1"].to(args.device)
                                    hidden_1 = result_2["hidden"].to(args.device)
                                    mean_1 = result_2["mean_1"].to(args.device)
                                    logvar_1 = result_2["logvar_1"].to(args.device)

                                    loss1, kl_divergence_z, atten_loss1 = self.return_loss(X1, X1_raw, latent_z1,
                                                                                           mean_1, logvar_1, hidden_1,
                                                                                           size_factor1, criterion,
                                                                                           attention_loss)
                                    test_loss = torch.mean(
                                        loss1 + (kl_weight * kl_divergence_z) + (args.sf2 * (atten_loss1)))

                        train_loss_list.append(test_loss.item())

                        print(
                            str(epoch) + "   " + str(test_loss.item()) + "   " + str(torch.mean(loss1).item()) + "   " +
                            str(torch.mean(kl_divergence_z).item()) + "   " + str(torch.mean(atten_loss1).item()))

                        if math.isnan(test_loss.item()):
                            flag_break = 1
                            break

                        if test_like_max > test_loss.item():
                            test_like_max = test_loss.item()
                            reco_epoch_test = epoch
                            patience_epoch = 0

            if flag_break == 1:
                print("containin NA")
                print(epoch)
                break

            if patience_epoch >= 30:
                print("patient with 30")
                print(epoch)
                break

            if len(train_loss_list) >= 2:
                if abs(train_loss_list[-1] - train_loss_list[-2]) / train_loss_list[-2] < 1e-4:
                    print("converged!!!")
                    print(epoch)
                    break

        duration = time.time() - start

        print('Finish training, total time is: ' + str(duration) + 's')
        self.eval()
        print(self.training)

        print('train likelihood is :  ' + str(test_like_max) + ' epoch: ' + str(reco_epoch_test))


class DCCA(nn.Module):
    """DCCA class.

    Parameters
    ----------
    layer_e_1 : list[int]
        Hidden layer specification for encoder1. List the dimensions of each hidden layer sequentially.
    hidden1_1 : int
        Hidden dimension for encoder1. It should be consistent with the last layer in layer_e_1.
    Zdim_1 : int
        Latent space dimension for VAE1.
    layer_d_1 : list[int]
        Hidden layer specification for decoder1. List the dimensions of each hidden layer sequentially.
    hidden2_1 : int
        Hidden dimension for decoder1. It should be consistent with the last layer in layer_d_1.
    layer_e_2 : int
        Hidden layer specification for encoder2. List the dimensions of each hidden layer sequentially.
    hidden1_2 : int
        Hidden dimension for encoder2. It should be consistent with the last layer in layer_e_1.
    Zdim_2 : int
        Latent space dimension for VAE2.
    layer_d_2 : int
        Hidden layer specification for decoder2. List the dimensions of each hidden layer sequentially.
    hidden2_2 : int
        Hidden dimension for decoder2. It should be consistent with the last layer in layer_d_1.
    args : argparse.Namespace
        A Namespace object that contains arguments of DCCA. For details of parameters in parser args, please refer to link (parser help document).
    ground_truth1 : pandas.Series
        Extra labels for VAE1.
    ground_truth2 :
        Extra labels for VAE2.
    Type_1 : str optional
        Loss type for VAE1. Default: 'NB'. By default to be 'NB'.
    Type_2 : str optional
        Loss type for VAE2. Default: 'Bernoulli'. By default to be 'Bernoulli'.
    cycle : int optional
        Number of multiple training cycles. In each cycle iteratively update VAE1 and VAE2. By default to be 1.
    attention_loss : str optional
        Loss type of attention loss. By default to be 'Eucli'.
    droprate : float optional
        Dropout rate for encoder/decoder layers. By default to be 0.1.
    """

    def __init__(self, layer_e_1, hidden1_1, Zdim_1, layer_d_1, hidden2_1,
                 layer_e_2, hidden1_2, Zdim_2, layer_d_2, hidden2_2, args,
                 ground_truth1, ground_truth2, Type_1='NB', Type_2='Bernoulli', cycle=1,
                 attention_loss='Eucli', droprate=0.1):

        super(DCCA, self).__init__()
        # cycle indicates the mutual learning, 0 for initiation of model1 with scRNA-seq data,
        # and odd for training other models, even for scRNA-seq

        self.model1 = VAE(layer_e=layer_e_1, hidden1=hidden1_1, Zdim=Zdim_1,
                          layer_d=layer_d_1, hidden2=hidden2_1,
                          Type=Type_1, droprate=droprate).to(args.device)
        self.model2 = VAE(layer_e=layer_e_2, hidden1=hidden1_2, Zdim=Zdim_2,
                          layer_d=layer_d_2, hidden2=hidden2_2,
                          Type=Type_2, droprate=droprate).to(args.device)

        if attention_loss == 'NST':
            self.attention = NSTLoss()

        elif attention_loss == 'FT':
            self.attention = FactorTransfer()

        elif attention_loss == 'SL':
            self.attention = Similarity()

        elif attention_loss == 'CC':
            self.attention = Correlation()

        elif attention_loss == 'AT':
            self.attention = Attention()

        elif attention_loss == 'KL_div':
            self.attention = KL_diver()

        elif attention_loss == 'L1':
            self.attention = L1_dis()

        else:
            self.attention = Eucli_dis()

        self.cycle = cycle
        self.args = args
        self.ground_truth1 = ground_truth1
        self.ground_truth2 = ground_truth2
        self.attention_loss = attention_loss

    def fit(self, train_loader, test_loader, total_loader, first="RNA"):
        """fit function for training

        Parameters
        ----------
        train_loader : torch.utils.data.DataLoader
            Dataloader for training dataset.
        test_loader : torch.utils.data.DataLoader
            Dataloader for testing dataset.
        total_loader : torch.utils.data.DataLoader
            Dataloader for both training and testing dataset, for extra evaluation purpose.
        first : str
            Type of modality 1.

        Returns
        -------
        None.
        """

        used_cycle = 0

        if self.ground_truth1 is not None:
            self.score(total_loader)

        while used_cycle < (self.cycle + 1):

            if first == "RNA":

                if used_cycle % 2 == 0:

                    self.model2.eval()

                    if used_cycle == 0:

                        self.model1.fit(train_loader, test_loader, total_loader, self.model2, self.args, self.attention,
                                        used_cycle, 0, first, self.attention_loss)

                    else:
                        self.model1.fit(train_loader, test_loader, total_loader, self.model2, self.args, self.attention,
                                        used_cycle, 1, first, self.attention_loss)

                else:
                    self.model1.eval()

                    if used_cycle == 1:

                        self.model2.fit(train_loader, test_loader, total_loader, self.model1, self.args, self.attention,
                                        used_cycle, 0, first, self.attention_loss)

                        if self.ground_truth1 is not None:
                            self.score(total_loader)

                        if self.attention_loss is not None:
                            self.model2.fit(train_loader, test_loader, total_loader, self.model1, self.args,
                                            self.attention, used_cycle, 1, first, self.attention_loss)

                    else:
                        self.model2.fit(train_loader, test_loader, total_loader, self.model1, self.args, self.attention,
                                        used_cycle, 1, first, self.attention_loss)

            else:
                if used_cycle % 2 == 0:

                    self.model1.eval()

                    if used_cycle == 0:

                        self.model2.fit(train_loader, test_loader, total_loader, self.model1, self.args, self.attention,
                                        used_cycle, 0, first, self.attention_loss)

                    else:
                        self.model2.fit(train_loader, test_loader, total_loader, self.model1, self.args, self.attention,
                                        used_cycle, 1, first, self.attention_loss)

                else:
                    self.model2.eval()

                    if used_cycle == 1:

                        self.model1.fit(train_loader, test_loader, total_loader, self.model2, self.args, self.attention,
                                        used_cycle, 0, first, self.attention_loss)

                        if self.ground_truth1 is not None:
                            self.score(total_loader)

                        self.model1.fit(train_loader, test_loader, total_loader, self.model2, self.args, self.attention,
                                        used_cycle, 1, first, self.attention_loss)

                    else:
                        self.model1.fit(train_loader, test_loader, total_loader, self.model2, self.args, self.attention,
                                        used_cycle, 1, first, self.attention_loss)

            used_cycle = used_cycle + 1

    def score(self, dataloader):
        """score function to get score of prediction.

         Parameters
         ----------
         dataloader : torch.utils.data.DataLoader
             Dataloader for testing dataset.

         Returns
         -------
         NMI_score1 : float
             Metric eval score for VAE1.
         ARI_score1 : float
             Metric eval score for VAE1.
         NMI_score2 : float
             Metric eval score for VAE2.
         ARI_score2 : float
             Metric eval score for VAE2.
         """

        self.model1.eval()
        self.model2.eval()

        with torch.no_grad():

            kmeans1 = KMeans(n_clusters=self.args.cluster1, n_init=5, random_state=200)
            kmeans2 = KMeans(n_clusters=self.args.cluster2, n_init=5, random_state=200)

            latent_code_rna = []
            latent_code_atac = []

            ARI_score1, NMI_score1 = -100, -100
            ARI_score2, NMI_score2 = -100, -100

            for batch_idx, (X1, _, size_factor1, X2, _, size_factor2) in enumerate(dataloader):


                X1, size_factor1 = X1.to(self.args.device), size_factor1.to(self.args.device)
                X2, size_factor2 = X2.to(self.args.device), size_factor2.to(self.args.device)

                X1, size_factor1 = Variable(X1), Variable(size_factor1)
                X2, size_factor2 = Variable(X2), Variable(size_factor2)

                result1 = self.model1.inference(X1, size_factor1)
                result2 = self.model2.inference(X2, size_factor2)

                latent_code_rna.append(result1["latent_z1"].data.cpu().numpy())
                latent_code_atac.append(result2["latent_z1"].data.cpu().numpy())

            latent_code_rna = np.concatenate(latent_code_rna)
            latent_code_atac = np.concatenate(latent_code_atac)

            pred_z1 = kmeans1.fit_predict(latent_code_rna)
            NMI_score1 = round(normalized_mutual_info_score(self.ground_truth1, pred_z1, average_method='max'), 3)
            ARI_score1 = round(metrics.adjusted_rand_score(self.ground_truth1, pred_z1), 3)

            pred_z2 = kmeans1.fit_predict(latent_code_atac)
            NMI_score2 = round(normalized_mutual_info_score(self.ground_truth1, pred_z2, average_method='max'), 3)
            ARI_score2 = round(metrics.adjusted_rand_score(self.ground_truth1, pred_z2), 3)

            print('scRNA-ARI: ' + str(ARI_score1) + ' NMI: ' + str(NMI_score1) + ' scEpigenomics-ARI: ' + str(
                ARI_score2) + ' NMI: ' + str(NMI_score2))
            return NMI_score1, ARI_score1, NMI_score2, ARI_score2

    def _encodeBatch(self, total_loader):
        """Helper function to get latent representation, normalized representation and prediction of data.

        Parameters
        ----------
        total_loader : torch.utils.data.DataLoader
            Dataloader for dataset.

        Returns
        -------
        latent_z1 : numpy.ndarray
            Latent representation of modality 1.
        latent_z2 : numpy.ndarray
            Latent representation of modality 2.
        norm_x1 : numpy.ndarray
            Normalized representation of modality 1.
        recon_x1 : numpy.ndarray
            Reconstruction result of modality 1.
        norm_x2 : numpy.ndarray
            Normalized representation of modality 2.
        recon_x2 : numpy.ndarray
            Reconstruction result of modality 2.
        """

        # processing large-scale datasets
        latent_z1 = []
        latent_z2 = []
        norm_x1 = []
        recon_x1 = []
        norm_x2 = []
        recon_x2 = []

        for batch_idx, (X1, _, size_factor1, X2, _, size_factor2) in enumerate(total_loader):
            X1, size_factor1 = X1.to(self.args.device), size_factor1.to(self.args.device)
            X2, size_factor2 = X2.to(self.args.device), size_factor2.to(self.args.device)

            X1, size_factor1 = Variable(X1), Variable(size_factor1)
            X2, size_factor2 = Variable(X2), Variable(size_factor2)

            result1 = self.model1(X1, size_factor1)
            result2 = self.model2(X2, size_factor2)

            latent_z1.append(result1["latent_z1"].data.cpu().numpy())
            latent_z2.append(result2["latent_z1"].data.cpu().numpy())

            norm_x1.append(result1["norm_x"].data.cpu().numpy())
            recon_x1.append(result1["recon_x"].data.cpu().numpy())

            norm_x2.append(result2["norm_x"].data.cpu().numpy())
            recon_x2.append(result2["recon_x"].data.cpu().numpy())

        latent_z1 = np.concatenate(latent_z1)
        latent_z2 = np.concatenate(latent_z2)
        norm_x1 = np.concatenate(norm_x1)
        recon_x1 = np.concatenate(recon_x1)
        norm_x2 = np.concatenate(norm_x2)
        recon_x2 = np.concatenate(recon_x2)

        return latent_z1, latent_z2, norm_x1, recon_x1, norm_x2, recon_x2

    def forward(self, total_loader):
        """Forward function for torch.nn.Module. An alias of encode_Batch function.

        Parameters
        ----------
        total_loader : torch.utils.data.DataLoader
            Dataloader for dataset.

        Returns
        -------
        latent_z1 : numpy.ndarray
            Latent representation of modality 1.
        latent_z2 : numpy.ndarray
            Latent representation of modality 2.
        norm_x1 : numpy.ndarray
            Normalized representation of modality 1.
        recon_x1 : numpy.ndarray
            Reconstruction result of modality 1.
        norm_x2 : numpy.ndarray
            Normalized representation of modality 2.
        recon_x2 : numpy.ndarray
            Reconstruction result of modality 2.
        """

        latent_z1, latent_z2, norm_x1, recon_x1, norm_x2, recon_x2 = self._encodeBatch(total_loader)

        return latent_z1, latent_z2, norm_x1, recon_x1, norm_x2, recon_x2

    def predict(self, total_loader):
        """Predict function to get latent representation of data.

        Parameters
        ----------
        total_loader : torch.utils.data.DataLoader
            Dataloader for dataset.

        Returns
        -------
        emb1 : numpy.ndarray
            Latent representation of modality 1.
        emb2 : numpy.ndarray
            Latent representation of modality 2.
        """
        self.eval()
        with torch.no_grad():
            emb1, emb2, _, _, _, _ = self.forward(total_loader)

        return emb1, emb2


import numpy as np
from sklearn.decomposition import NMF
from sklearn.linear_model import LinearRegression

def cell_topic_profile(X, groups, axis=1, method='median'):
    """cell_topic_profile.

    Parameters
    ----------
    X : numpy 2-d array
        gene expression matrix, genes (rows) by sample cells (cols).
    groups : int
        cell-type labels of each sample cell in X.
    method : string optional
         method for reduction of cell-types for cell profile, default median.
           
    Returns
        -------
    X_profile : numpy 2-d array
         cell profile matrix from scRNA-seq reference.

    """
    groups = np.array(groups)
    ids=np.unique(groups)
    if method == "median":
        X_profile = np.array([np.median(X[:, groups==ids[i]], axis=1) for i in range(len(ids))]).T
    else:
        X_profile = np.array([np.mean(X[:, groups==ids[i]], axis=1) for i in range(len(ids))]).T
    return X_profile
    
    
class SPOTlight(object):    
    """SPOTlight class.

    Parameters
    ----------
    rank : int
           rank of the matrix factorization.
    bias : boolean optional
           set true to include bias term in the regression modules, default False.
    profile_mtd : string optional
           method for reduction of cell-types for cell profile, default median.
    init : string optional
           initialization method for matrix factorization solver (see NMF from sklearn).
    max_iter : int optional
           maximum iterations allowed for matrix factorization solver.
    
    Returns
        -------
    None.

    """
    def __init__(self, rank, bias=False, profile_mtd='median', init='random', random_state=0, max_iter=1000):
        super(SPOTlight, self).__init__()
        self.bias = bias
        self.profile_mtd = 'median'
        
        self.nmf_model = NMF(n_components=rank, init=init,  random_state=random_state, max_iter=max_iter)

        self.nnls_reg1=LinearRegression(fit_intercept=bias,positive=True)

        self.nnls_reg2=LinearRegression(fit_intercept=bias,positive=True)
    
    def fit(self,x,y, cell_types):
        """fit function for model training.

        Parameters
        ----------
        x : 
            scRNA-seq reference expression.
        y : 
            mixed cell expression to be deconvoluted.
        cell_types : 
            cell type annotations for scRNA-seq reference.
            
        Returns
        -------
        None.

        """
        #1. run NMF on scRNA X
        self.W = self.nmf_model.fit_transform(x)
        self.H = self.nmf_model.components_
        
        #2. get cell-topic and mix-topic profiles
        #a. get cell-topic profiles H_profile: cell-type group medians of coef H (topic x cells)
        self.H_profile = cell_topic_profile(self.H, groups=cell_types,axis=1, method=self.profile_mtd)
        
        #b. get mix-topic profiles B: NNLS of basis W onto mix expression Y -- y ~ W*b
        #nnls ran for each spot
        self.nnls_reg1.fit(self.W, y)
        self.B=self.nnls_reg1.coef_.T
        
        #3.  get cell-type proportions P: NNLS of cell-topic profile H_profile onoto mix-topic profile B -- b ~ h_profile*p 
        self.nnls_reg2.fit(self.H_profile, self.B)
        self.P=self.nnls_reg2.coef_.T
        
    def predict(self, y):
        """prediction function.
        Parameters
        ----------
        y : numpy 2-d array
            mixed cell expression.
        
        Returns
        -------
        P : numpy 2-d array
            predictions of cell-type proportions.
        
        """
        nnls = LinearRegression(fit_intercept=self.bias,positive=True)
        nnls.fit(self.W, y)
        B=nnls.coef_.T
        
        nnls = LinearRegression(fit_intercept=self.bias,positive=True)
        nnls.fit(self.H_profile, B)
        P=nnls.coef_.T
        return(P)
    
    def score(self, x, y, cell_types):
        """score.

        Parameters
        ----------
        x : 
            scRNA-seq reference expression.
        y : 
            mixed cell expression to be deconvoluted.
        cell_types : 
            cell type annotations for scRNA-seq reference.

        Returns
        -------
        model_score : float
            coefficient of determination of the prediction (final non-negative linear module).

        """
        self.fit(x,y,cell_types)
        model_score = self.nnls_reg2.score(self.H_profile, self.B)
        return model_score
        
        


import csv
import glob
import os
import os.path as osp
import pickle as pkl
import random
import re
import time as tm
import warnings
from collections import defaultdict
from operator import itemgetter

import anndata
import cv2
import networkx as nx
import numpy as np
import pandas as pd
import rdata
import scanpy as sc
import scipy.sparse
from anndata import AnnData
from scipy.stats import uniform

from dance.data import download_file, download_unzip, unzip_file
from dance.transforms import preprocess

IGNORED_FILES = ["readme.txt"]

dataset = {
    "151510": "https://www.dropbox.com/sh/41h9brsk6my546x/AADa18mkJge-KQRTndRelTpMa?dl=1",
    "151507": "https://www.dropbox.com/sh/m3554vfrdzbwv2c/AACGsFNVKx8rjBgvF7Pcm2L7a?dl=1",
    "151508": "https://www.dropbox.com/sh/tm47u3fre8692zt/AAAJJf8-za_Lpw614ft096qqa?dl=1",
    "151509": "https://www.dropbox.com/sh/hihr7906vyirjet/AACslV5mKIkF2CF5QqE1LE6ya?dl=1",
    "151669": "https://www.dropbox.com/sh/ulw2nnnmgtbswvc/AAC0fT549EwtxKZWWoB89gb4a?dl=1",
    "151670": "https://www.dropbox.com/sh/8fw44zyyjgh0ddc/AAA1asGAmyDiMmvhRmL7pN1Na?dl=1",
    "151671": "https://www.dropbox.com/sh/9g5qzd5ykx2mpk3/AAD3xjx1i2h0RhYBc-Vft6CEa?dl=1",
    "151672": "https://www.dropbox.com/sh/l6519tr280krd4p/AAAWefCSp2iKhVmLgytlyxTta?dl=1",
    "151673": "https://www.dropbox.com/sh/qc64ps6gd64dm0c/AAC_5_mP4AczKj8lORLLKcIba?dl=1",
    "151674": "https://www.dropbox.com/sh/q7io99psd2xuqgw/AABske8dgX_kc1oaDSxuiqjpa?dl=1",
    "151675": "https://www.dropbox.com/sh/uahka2h5klnrzvj/AABe7K0_ewqOcqKUxHebE6qLa?dl=1",
    "151676": "https://www.dropbox.com/sh/jos5jjurezy5zp1/AAB2uaVm3-Us1a4mDkS1Q-iAa?dl=1",
}

cellDeconvo_dataset = {
    "CARD_synthetic": "https://www.dropbox.com/sh/v0vpv0jsnfexj7f/AADpizLGOrF7M8EesDihgbBla?dl=1",
    "GSE174746": "https://www.dropbox.com/sh/spfv06yfttetrab/AAAgORS6ocyoZEyxiRYKTymCa?dl=1",
    "SPOTLight_synthetic": "https://www.dropbox.com/sh/p1tfb0xe1yl2zpe/AAB6cF-BsdJcHToet_C-AlXAa?dl=1",
    "human PDAC": "https://www.dropbox.com/sh/9py6hk9j1ygyprh/AAAOKTo-TE_eX4JJg0HIFfZ7a?dl=1",
    "mouse brain 1": "https://www.dropbox.com/sh/e2nl247v1jrd7h8/AAC1IUlk_3vXUvfk2fv9L2D3a?dl=1",
    "toy1": "https://www.dropbox.com/sh/quvjz6pzltio43u/AAC8vd8-H-4S58-b1pGz3DLRa?dl=1",
    "toy2": "https://www.dropbox.com/sh/eqkcm344p5d1akr/AAAPs0Z0S7yFC5ML8Kcd5eU9a?dl=1",
}

card_simulation_dataset = {
    "sim_noise0_rep1": "https://www.dropbox.com/s/aujusznwbq4xa99/sim.pseudo.MOB.n10.cellType6.Mixnoise0.repeat1.RData?dl=1",
    "sim_noise0_rep2": "https://www.dropbox.com/s/bjyagojp3opkyxd/sim.pseudo.MOB.n10.cellType6.Mixnoise0.repeat2.RData?dl=1",
    "sim_noise0_rep3": "https://www.dropbox.com/s/k9n07ujvqtroj72/sim.pseudo.MOB.n10.cellType6.Mixnoise0.repeat3.RData?dl=1",
    "sim_noise0_rep4": "https://www.dropbox.com/s/3m49dq387he776x/sim.pseudo.MOB.n10.cellType6.Mixnoise0.repeat4.RData?dl=1",
    "sim_noise0_rep5": "https://www.dropbox.com/s/4wjsl2ids1q16b2/sim.pseudo.MOB.n10.cellType6.Mixnoise0.repeat5.RData?dl=1",
    "sim_noise3_rep1": "https://www.dropbox.com/s/z6ehj48q0vcxf13/sim.pseudo.MOB.n10.cellType6.Mixnoise3.repeat1.RData?dl=1",
    "sim_noise3_rep2": "https://www.dropbox.com/s/2mzjt4pv1f5ucs2/sim.pseudo.MOB.n10.cellType6.Mixnoise3.repeat2.RData?dl=1",
    "sim_noise3_rep3": "https://www.dropbox.com/s/65ixbnu6x65o8ee/sim.pseudo.MOB.n10.cellType6.Mixnoise3.repeat3.RData?dl=1",
    "sim_noise3_rep4": "https://www.dropbox.com/s/hrmwoi14wta0ida/sim.pseudo.MOB.n10.cellType6.Mixnoise3.repeat4.RData?dl=1",
    "sim_noise3_rep5": "https://www.dropbox.com/s/0txpltfj2p3dz9v/sim.pseudo.MOB.n10.cellType6.Mixnoise3.repeat5.RData?dl=1",
}


class SpotDataset:
    def __init__(self, data_id="151673", data_dir="data/spot", build_graph_fn="default"):
        self.data_id = data_id
        self.data_dir = data_dir + "/{}".format(data_id)
        self.data_url = dataset[data_id]
        self.load_data()
        self.adj = None

    def get_all_data(self):
        # provide an interface to get all data at one time
        print("All data includes {} datasets: {}".format(len(dataset), ",".join(dataset.keys())))
        res = {}
        for each_dataset in dataset.keys():
            res[each_dataset] = SpotDataset(each_dataset)
        return res

    def download_data(self):
        # judge whether a file exists or not
        isdownload = download_file(self.data_url, self.data_dir + "/{}.zip".format(self.data_id))
        if isdownload:
            unzip_file(self.data_dir + "/{}.zip".format(self.data_id), self.data_dir + "/")
        return self

    def is_complete(self):
        # data.h5ad
        # histology.tif
        # positions.txt
        # judge whether data is complete or not
        check = [
            self.data_dir + "/{}_raw_feature_bc_matrix.h5".format(self.data_id),
            self.data_dir + "/{}_full_image.tif".format(self.data_id), self.data_dir + "/tissue_positions_list.txt"
        ]

        for i in check:
            if not os.path.exists(i):
                print("lack {}".format(i))
                return False
        return True

    def load_data(self):
        if self.is_complete():
            pass
        else:
            self.download_data()
        self.data = sc.read_10x_h5(self.data_dir + "/{}_raw_feature_bc_matrix.h5".format(self.data_id))
        self.img = cv2.imread(self.data_dir + "/{}_full_image.tif".format(self.data_id))
        label = pd.read_csv(self.data_dir + "/cluster_labels.csv")
        classes = {layer_class: idx for idx, layer_class in enumerate(set(label["ground_truth"].tolist()))}
        self.spatial = pd.read_csv(self.data_dir + "/tissue_positions_list.txt", sep=",", header=None, na_filter=False,
                                   index_col=0)
        self.data.obs["x1"] = self.spatial[1]
        self.data.obs["x2"] = self.spatial[2]
        self.data.obs["x3"] = self.spatial[3]
        self.data.obs["x4"] = self.spatial[4]
        self.data.obs["x5"] = self.spatial[5]
        self.data.obs["x"] = self.data.obs["x2"]
        self.data.obs["y"] = self.data.obs["x3"]
        self.data.obs["x_pixel"] = self.data.obs["x4"]
        self.data.obs["y_pixel"] = self.data.obs["x5"]

        self.data = self.data[self.data.obs["x1"] == 1]
        self.data.var_names = [i.upper() for i in list(self.data.var_names)]
        self.data.var["genename"] = self.data.var.index.astype("str")
        self.data.obs["label"] = list(map(lambda x: classes[x], label["ground_truth"].tolist()))
        self.data.obs["ground_truth"] = label["ground_truth"].tolist()
        return self


class CellTypeDeconvoDataset:
    def __init__(self, data_id="toy1", data_dir="data/spatial", build_graph_fn="default"):
        self.data_id = data_id
        self.data_dir = data_dir + "/{}".format(data_id)
        self.data_url = cellDeconvo_dataset[data_id]
        self.load_data()
        self.adj = None

    def get_all_data(self):
        # provide an interface to get all data at one time
        print("All data includes {} cellDeconvo datasets: {}".format(len(cellDeconvo_dataset),
                                                                     ",".join(cellDeconvo_dataset.keys())))
        res = {}
        for each_dataset in cellDeconvo_dataset.keys():
            res[each_dataset] = cellDeconvo_dataset(each_dataset)
        return res

    def download_data(self):
        # judge whether a file exists or not
        isdownload = download_file(self.data_url, self.data_dir + "/{}.zip".format(self.data_id))
        if isdownload:
            unzip_file(self.data_dir + "/{}.zip".format(self.data_id), self.data_dir)
        return self

    def is_complete(self):
        check = [self.data_dir + "/mix_count.*", self.data_dir + "/ref_sc_count.*"]

        for i in check:
            #if not os.path.exists(i):
            if not glob.glob(i):
                print("lack {}".format(i))
                return False
        return True

    def load_data(self):
        if self.is_complete():
            pass
        else:
            self.download_data()

        self.data = {}
        files = os.listdir(self.data_dir + "/")
        filenames = [f.split(".")[0] for f in files]
        extensions = [f.split(".")[1] for f in files]
        for f in files:
            DataPath = self.data_dir + "/" + f
            filename = f.split(".")[0]
            ext = f.split(".")[1]
            if ext == "csv":
                data = pd.read_csv(DataPath, header=0, index_col=0)
                self.data[filename] = data
            elif ext == "h5ad":
                data = sc.read_h5ad(DataPath)
                self.data[filename] = data
                self.data[filename + "_annot"] = data.obs
            else:
                print("unsupported file type. Please use csv or h5ad file types.")

        print("load data succesfully....")

        return self


class CellTypeDeconvoDatasetLite:
    def __init__(self, data_id="GSE174746", data_dir="data/spatial", build_graph_fn="default"):
        if data_id not in cellDeconvo_dataset:
            raise ValueError(f"Unknown data_id {data_id!r}, available datasets are: {sorted(cellDeconvo_dataset)}")

        self.data_id = data_id
        self.data_dir = osp.join(data_dir, data_id)
        self.data_url = cellDeconvo_dataset[data_id]
        self.load_data()

    def load_data(self):
        if not osp.exists(self.data_dir):
            download_unzip(self.data_url, self.data_dir)

        self.data = {}
        for f in os.listdir(self.data_dir):
            filepath = osp.join(self.data_dir, f)
            filename, ext = osp.splitext(f)
            if f in IGNORED_FILES:
                continue
            elif ext == ".csv":
                self.data[filename] = pd.read_csv(filepath, header=0, index_col=0)
            elif ext == ".h5ad":
                self.data[filename] = sc.read_h5ad(filepath).to_df()
            else:
                warnings.warn(f"Unsupported file type {ext!r}. Use csv or h5ad file types.")


class CARDSimulationRDataset:
    ref_sc_count_url: str = "https://www.dropbox.com/s/wchoppxcsulk8ev/split2_ref_sc_count.h5ad?dl=1"
    ref_sc_annot_url: str = "https://www.dropbox.com/s/irpvco2ffisvxvk/split2_ref_sc_annot.csv?dl=1"

    def __init__(self, data_id="sim_noise0_rep1", data_dir="data/spatial/card_simulation", build_graph_fn="default"):
        self.data_id = data_id
        self.data_dir = osp.join(data_dir, data_id)
        self.data_path = osp.join(data_dir, f"{data_id}.RData")
        self.ref_sc_count_path = osp.join(data_dir, "ref_sc_count.h5ad")
        self.ref_sc_annot_path = osp.join(data_dir, "ref_sc_annot.csv")
        self.data_url = card_simulation_dataset[data_id]
        self.load_data()
        self.adj = None

    def get_all_data(self):
        # TODO: make classmethod, make data url dict as class attrs
        dataset_info = "\n\t".join(list(card_simulation_dataset))
        print(f"Total of {(len(card_simulation_dataset))} datasets:\n{dataset_info}")
        return {i: CARDSimulationRDataset(i) for i in card_simulation_dataset}

    def download_data(self):
        download_file(self.ref_sc_count_url, self.ref_sc_count_path)
        download_file(self.ref_sc_annot_url, self.ref_sc_annot_path)
        download_file(self.data_url, self.data_path)

    def is_complete(self):
        check = [self.data_path, self.ref_sc_count_path, self.ref_sc_annot_path]
        return all(map(osp.exists, check))

    def load_data(self):
        if not self.is_complete():
            self.download_data()

        raw = rdata.conversion.convert(rdata.parser.parse_file(self.data_path))["spatial.pseudo"]
        ref_sc_count = anndata.read_h5ad(self.ref_sc_count_path).to_df().T
        ref_sc_annot = pd.read_csv(self.ref_sc_annot_path, index_col=0)
        spatial_count = raw["pseudo.data"].to_pandas().T
        spatial_location = (
            spatial_count
            .reset_index()["dim_1"]
            .str
            .split("x", expand=True)
            .set_index(spatial_count.index)
            .rename({0: "x", 1: "y"}, axis=1)
            .astype(float)
        )
        true_p = raw["true.p"].to_pandas()

        # TODO: directly save as attrs instead of a dict?
        self.data = {
            "ref_sc_count": ref_sc_count,
            "ref_sc_annot": ref_sc_annot,
            "spatial_count": spatial_count,
            "spatial_location": spatial_location,
            "true_p": true_p,
        }

        return self



"""
This is the implementation of scDCC.
Extended from https://github.com/ttgump/scDCC

Reference
----------
Tian, Tian, et al. "Model-based deep embedding for constrained clustering analysis of single cell RNA-seq data." Nature communications 12.1 (2021): 1-12.
"""

import math, os
import numpy as np
from sklearn.cluster import KMeans
from sklearn import metrics

import torch
import torch.nn as nn
from torch.autograd import Variable
from torch.nn import Parameter
import torch.nn.functional as F
import torch.optim as optim
from torch.utils.data import DataLoader, TensorDataset

from dance.utils.metrics import cluster_acc
from dance.utils.loss import ZINBLoss


def buildNetwork(layers, type, activation="relu"):
    """Build network layer.

    Parameters
    ----------
    layers : list
        dimensions of layers.
    type : str
        type of network.
    activation : str optional
        activation function.

    Returns
    -------
    net :
        torch.nn network.

    """
    net = []
    for i in range(1, len(layers)):
        net.append(nn.Linear(layers[i - 1], layers[i]))
        if activation == "relu":
            net.append(nn.ReLU())
        elif activation == "sigmoid":
            net.append(nn.Sigmoid())
    net = nn.Sequential(*net)
    return net


class ScDCC(nn.Module):
    """scDCC class.

    Parameters
    ----------
    input_dim : int
        dimension of encoder input.
    z_dim : int
        dimension of embedding.
    n_clusters : int
        number of clusters.
    encodeLayer : list optional
        dimensions of encoder layers.
    decodeLayer : list optional
        dimensions of decoder layers.
    activation : str optional
        activation function.
    sigma : float optional
        parameter of Gaussian noise.
    alpha : float optional
        parameter of soft assign.
    gamma : float optional
        parameter of cluster loss.
    ml_weight : float optional
        parameter of must-link loss.
    cl_weight : float optional
        parameter of cannot-link loss.

    """

    def __init__(self, input_dim, z_dim, n_clusters, encodeLayer=[], decodeLayer=[],
                 activation="relu", sigma=1., alpha=1., gamma=1., ml_weight=1., cl_weight=1.):
        super(ScDCC, self).__init__()
        self.z_dim = z_dim
        self.n_clusters = n_clusters
        self.activation = activation
        self.sigma = sigma
        self.alpha = alpha
        self.gamma = gamma
        self.ml_weight = ml_weight
        self.cl_weight = cl_weight
        self.encoder = buildNetwork([input_dim] + encodeLayer, type="encode", activation=activation)
        self.decoder = buildNetwork([z_dim] + decodeLayer, type="decode", activation=activation)
        self._enc_mu = nn.Linear(encodeLayer[-1], z_dim)
        self._dec_mean = nn.Sequential(nn.Linear(decodeLayer[-1], input_dim), MeanAct())
        self._dec_disp = nn.Sequential(nn.Linear(decodeLayer[-1], input_dim), DispAct())
        self._dec_pi = nn.Sequential(nn.Linear(decodeLayer[-1], input_dim), nn.Sigmoid())
        self.mu = Parameter(torch.Tensor(n_clusters, z_dim))
        self.zinb_loss = ZINBLoss().cpu()

    def save_model(self, path):
        """Save model to path.

        Parameters
        ----------
        path : str
            path to save model.

        Returns
        -------
        None.

        """
        torch.save(self.state_dict(), path)

    def load_model(self, path):
        """Load model from path.

        Parameters
        ----------
        path : str
            path to load model.

        Returns
        -------
        None.

        """
        pretrained_dict = torch.load(path, map_location=lambda storage, loc: storage)
        model_dict = self.state_dict()
        pretrained_dict = {k: v for k, v in pretrained_dict.items() if k in model_dict}
        model_dict.update(pretrained_dict)
        self.load_state_dict(model_dict)

    def soft_assign(self, z):
        """Soft assign q with z.

        Parameters
        ----------
        z :
            embedding.

        Returns
        -------
        q :
            soft label.

        """
        q = 1.0 / (1.0 + torch.sum((z.unsqueeze(1) - self.mu) ** 2, dim=2) / self.alpha)
        q = q ** ((self.alpha + 1.0) / 2.0)
        q = (q.t() / torch.sum(q, dim=1)).t()
        return q

    def target_distribution(self, q):
        """Calculate auxiliary target distribution p with q.

        Parameters
        ----------
        q :
            soft label.

        Returns
        -------
        p :
            target distribution.

        """
        p = q ** 2 / q.sum(0)
        return (p.t() / p.sum(1)).t()

    def forward(self, x):
        """Forward propagation.

        Parameters
        ----------
        x :
            input features.

        Returns
        -------
        z0 :
            embedding.
        q :
            soft label.
        _mean :
            data mean from ZINB.
        _disp :
            data dispersion from ZINB.
        _pi :
            data dropout probability from ZINB.

        """
        h = self.encoder(x + torch.randn_like(x) * self.sigma)
        z = self._enc_mu(h)
        h = self.decoder(z)
        _mean = self._dec_mean(h)
        _disp = self._dec_disp(h)
        _pi = self._dec_pi(h)

        h0 = self.encoder(x)
        z0 = self._enc_mu(h0)
        q = self.soft_assign(z0)
        return z0, q, _mean, _disp, _pi

    def encodeBatch(self, X, batch_size=256):
        """Batch encoder.

        Parameters
        ----------
        X :
            input features.
        batch_size : int optional
            size of batch.

        Returns
        -------
        encoded :
            embedding.

        """
        use_cuda = torch.cuda.is_available()
        if use_cuda:
            device = 'cuda'
        else:
            device = 'cpu'
        self.to(device)

        encoded = []
        num = X.shape[0]
        num_batch = int(math.ceil(1.0 * X.shape[0] / batch_size))
        for batch_idx in range(num_batch):
            xbatch = X[batch_idx * batch_size: min((batch_idx + 1) * batch_size, num)]
            inputs = Variable(xbatch)
            z, _, _, _, _ = self.forward(inputs)
            encoded.append(z.data)

        encoded = torch.cat(encoded, dim=0)
        return encoded

    def cluster_loss(self, p, q):
        """Calculate cluster loss.

        Parameters
        ----------
        p :
            target distribution.
        q :
            soft label.

        Returns
        -------
        loss :
            cluster loss.

        """

        def kld(target, pred):
            return torch.mean(torch.sum(target * torch.log(target / (pred + 1e-6)), dim=-1))

        kldloss = kld(p, q)
        loss = self.gamma * kldloss
        return loss

    def pairwise_loss(self, p1, p2, cons_type):
        """Calculate pairwise loss.

        Parameters
        ----------
        p1 :
            distribution 1.
        p2 :
            distribution 2.
        cons_type : str
            type of loss.

        Returns
        -------
        loss :
            pairwise loss.

        """
        if cons_type == "ML":
            ml_loss = torch.mean(-torch.log(torch.sum(p1 * p2, dim=1)))
            loss = self.ml_weight * ml_loss
            return loss
        else:
            cl_loss = torch.mean(-torch.log(1.0 - torch.sum(p1 * p2, dim=1)))
            loss = self.cl_weight * cl_loss
            return loss

    def pretrain_autoencoder(self, x, X_raw, size_factor, batch_size=256, lr=0.001, epochs=400, ae_save=True,
                             ae_weights='AE_weights.pth.tar'):
        """Pretrain autoencoder.

        Parameters
        ----------
        x :
            input features.
        X_raw :
            raw input features.
        size_factor : list
            size factor of input features and raw input features.
        batch_size : int optional
            size of batch.
        lr : float optional
            learning rate.
        epochs : int optional
            number of epochs.
        ae_save : bool optional
            save autoencoder weights or not.
        ae_weights : str optional
            path to save autoencoder weights.

        Returns
        -------
        None.

        """
        use_cuda = torch.cuda.is_available()
        if use_cuda:
            device = 'cuda'
        else:
            device = 'cpu'
        self.to(device)
        dataset = TensorDataset(torch.Tensor(x), torch.Tensor(X_raw), torch.Tensor(size_factor))
        dataloader = DataLoader(dataset, batch_size=batch_size, shuffle=True)
        print("Pretraining stage")
        optimizer = optim.Adam(filter(lambda p: p.requires_grad, self.parameters()), lr=lr, amsgrad=True)
        for epoch in range(epochs):
            for batch_idx, (x_batch, x_raw_batch, sf_batch) in enumerate(dataloader):
                x_tensor = Variable(x_batch).to(device)
                x_raw_tensor = Variable(x_raw_batch).to(device)
                sf_tensor = Variable(sf_batch).to(device)
                _, _, mean_tensor, disp_tensor, pi_tensor = self.forward(x_tensor)
                loss = self.zinb_loss(x=x_raw_tensor, mean=mean_tensor, disp=disp_tensor, pi=pi_tensor,
                                      scale_factor=sf_tensor)
                optimizer.zero_grad()
                loss.backward()
                optimizer.step()
                print('Pretrain epoch [{}/{}], ZINB loss:{:.4f}'.format(batch_idx + 1, epoch + 1, loss.item()))

        if ae_save:
            torch.save({'ae_state_dict': self.state_dict(),
                        'optimizer_state_dict': optimizer.state_dict()}, ae_weights)

    def save_checkpoint(self, state, index, filename):
        """Save training checkpoint.

        Parameters
        ----------
        state :
            model state
        index : int
            checkpoint index
        filename : str
            filename to save

        Returns
        -------
        None.

        """
        newfilename = os.path.join(filename, 'FTcheckpoint_%d.pth.tar' % index)
        torch.save(state, newfilename)

    def fit(self, X, X_raw, sf, ml_ind1=np.array([]), ml_ind2=np.array([]), cl_ind1=np.array([]), cl_ind2=np.array([]),
            ml_p=1., cl_p=1., y=None, lr=1., batch_size=256, num_epochs=10, update_interval=1, tol=1e-3, save_dir=""):
        """Train model.

        Parameters
        ----------
        X :
            input features.
        X_raw :
            raw input features.
        sf : float
            size factor of input features and raw input features.
        ml_ind1 : np.array optional
            index 1 of must-link pairs.
        ml_ind2 : np.array optional
            index 2 of must-link pairs.
        cl_ind1 : np.array optional
            index 1 of cannot-link pairs.
        cl_ind2 : np.array optional
            index 2 of cannot-link pairs.
        ml_p : float optional
            parameter of must-link loss.
        cl_p : float optional
            parameter of cannot-link loss.
        y : list optional
            true label. Used for model selection.
        lr : float optional
            learning rate.
        batch_size : int optional
            size of batch.
        num_epochs : int optional
            number of epochs.
        update_interval : int optional
            update interval of soft label and target distribution.
        tol : float optional
            tolerance for training loss.
        save_dir : str optional
            path to save model weights.

        Returns
        -------
        None.

        """

        print("Training stage")
        use_cuda = torch.cuda.is_available()
        if use_cuda:
            device = 'cuda'
        else:
            device = 'cpu'
        self.to(device)
        X = torch.tensor(X).to(device)
        X_raw = torch.tensor(X_raw).to(device)
        sf = torch.tensor(sf).to(device)
        optimizer = optim.Adadelta(filter(lambda p: p.requires_grad, self.parameters()), lr=lr, rho=.95)

        # Initializing cluster centers with kmeans
        kmeans = KMeans(self.n_clusters, n_init=20)
        data = self.encodeBatch(X)
        self.y_pred = kmeans.fit_predict(data.data.cpu().numpy())
        self.y_pred_last = self.y_pred
        self.mu.data.copy_(torch.Tensor(kmeans.cluster_centers_))

        self.train()
        num = X.shape[0]
        num_batch = int(math.ceil(1.0 * X.shape[0] / batch_size))
        ml_num_batch = int(math.ceil(1.0 * ml_ind1.shape[0] / batch_size))
        cl_num_batch = int(math.ceil(1.0 * cl_ind1.shape[0] / batch_size))
        cl_num = cl_ind1.shape[0]
        ml_num = ml_ind1.shape[0]

        update_ml = 1
        update_cl = 1

        aris = []
        P = {}
        Q = {}

        for epoch in range(num_epochs):
            if epoch % update_interval == 0:
                # update the targe distribution p
                latent = self.encodeBatch(X)
                q = self.soft_assign(latent)
                self.q = q
                p = self.target_distribution(q).data
                self.y_pred = self.predict()

                # save current model
                if (epoch > 0 and delta_label < tol) or epoch % 10 == 0:
                    self.save_checkpoint({'epoch': epoch + 1,
                                          'state_dict': self.state_dict(),
                                          'mu': self.mu,
                                          'p': p,
                                          'q': q,
                                          'y': y
                                          }, epoch + 1, filename=save_dir)
                p_ = {f'epoch{epoch}': p}
                q_ = {f'epoch{epoch}': q}
                P = {**P, **p_}
                Q = {**Q, **q_}

                # check stop criterion
                delta_label = np.sum(self.y_pred != self.y_pred_last).astype(np.float32) / num
                self.y_pred_last = self.y_pred
                if epoch > 0 and delta_label < tol:
                    print('delta_label ', delta_label, '< tol ', tol)
                    print("Reach tolerance threshold. Stopping training.")
                    break

                # calculate ari score for model selection
                _, _, ari = self.score(y)
                aris.append(ari)

            # train 1 epoch for clustering loss
            train_loss = 0.0
            recon_loss_val = 0.0
            cluster_loss_val = 0.0
            for batch_idx in range(num_batch):
                xbatch = X[batch_idx * batch_size: min((batch_idx + 1) * batch_size, num)]
                xrawbatch = X_raw[batch_idx * batch_size: min((batch_idx + 1) * batch_size, num)]
                sfbatch = sf[batch_idx * batch_size: min((batch_idx + 1) * batch_size, num)]
                pbatch = p[batch_idx * batch_size: min((batch_idx + 1) * batch_size, num)]
                optimizer.zero_grad()
                inputs = Variable(xbatch)
                rawinputs = Variable(xrawbatch)
                sfinputs = Variable(sfbatch)
                target = Variable(pbatch)

                z, qbatch, meanbatch, dispbatch, pibatch = self.forward(inputs)

                cluster_loss = self.cluster_loss(target, qbatch)
                recon_loss = self.zinb_loss(rawinputs, meanbatch, dispbatch, pibatch, sfinputs)
                loss = cluster_loss + recon_loss
                loss.backward()
                optimizer.step()
                cluster_loss_val += cluster_loss.data * len(inputs)
                recon_loss_val += recon_loss.data * len(inputs)
                train_loss = cluster_loss_val + recon_loss_val

            print("#Epoch %3d: Total: %.4f, Clustering Loss: %.4f, ZINB Loss: %.4f" % (
                epoch + 1, train_loss / num, cluster_loss_val / num, recon_loss_val / num))

            ml_loss = 0.0
            if epoch % update_ml == 0:
                for ml_batch_idx in range(ml_num_batch):
                    px1 = X[ml_ind1[ml_batch_idx * batch_size: min(ml_num, (ml_batch_idx + 1) * batch_size)]]
                    pxraw1 = X_raw[ml_ind1[ml_batch_idx * batch_size: min(ml_num, (ml_batch_idx + 1) * batch_size)]]
                    sf1 = sf[ml_ind1[ml_batch_idx * batch_size: min(ml_num, (ml_batch_idx + 1) * batch_size)]]
                    px2 = X[ml_ind2[ml_batch_idx * batch_size: min(ml_num, (ml_batch_idx + 1) * batch_size)]]
                    sf2 = sf[ml_ind2[ml_batch_idx * batch_size: min(ml_num, (ml_batch_idx + 1) * batch_size)]]
                    pxraw2 = X_raw[ml_ind2[ml_batch_idx * batch_size: min(ml_num, (ml_batch_idx + 1) * batch_size)]]
                    optimizer.zero_grad()
                    inputs1 = Variable(px1)
                    rawinputs1 = Variable(pxraw1)
                    sfinput1 = Variable(sf1)
                    inputs2 = Variable(px2)
                    rawinputs2 = Variable(pxraw2)
                    sfinput2 = Variable(sf2)
                    z1, q1, mean1, disp1, pi1 = self.forward(inputs1)
                    z2, q2, mean2, disp2, pi2 = self.forward(inputs2)
                    loss = (ml_p * self.pairwise_loss(q1, q2, "ML") + self.zinb_loss(rawinputs1, mean1, disp1, pi1,
                                                                                     sfinput1) + self.zinb_loss(
                        rawinputs2, mean2, disp2, pi2, sfinput2))
                    # 0.1 for mnist/reuters, 1 for fashion, the parameters are tuned via grid search on validation set
                    ml_loss += loss.data
                    loss.backward()
                    optimizer.step()

            cl_loss = 0.0
            if epoch % update_cl == 0:
                for cl_batch_idx in range(cl_num_batch):
                    px1 = X[cl_ind1[cl_batch_idx * batch_size: min(cl_num, (cl_batch_idx + 1) * batch_size)]]
                    px2 = X[cl_ind2[cl_batch_idx * batch_size: min(cl_num, (cl_batch_idx + 1) * batch_size)]]
                    optimizer.zero_grad()
                    inputs1 = Variable(px1)
                    inputs2 = Variable(px2)
                    z1, q1, _, _, _ = self.forward(inputs1)
                    z2, q2, _, _, _ = self.forward(inputs2)
                    loss = cl_p * self.pairwise_loss(q1, q2, "CL")
                    cl_loss += loss.data
                    loss.backward()
                    optimizer.step()

            if ml_num_batch > 0 and cl_num_batch > 0:
                print("Pairwise Total: %.4f, ML loss: %.4f, CL loss: %.4f" % (
                    float(ml_loss.cpu()) + float(cl_loss.cpu()), ml_loss.cpu(), cl_loss.cpu()))

        index = update_interval * np.argmax(aris)
        self.q = Q[f'epoch{index}']

    def predict(self):
        """Get predictions from the trained model.

        Parameters
        ----------
        None.

        Returns
        -------
        y_pred : np.array
            prediction of given clustering method.

        """
        y_pred = torch.argmax(self.q, dim=1).data.cpu().numpy()
        return y_pred

    def score(self, y):
        """Evaluate the trained model.

        Parameters
        ----------
        y : list
            true labels.

        Returns
        -------
        acc : float
            accuracy.
        nmi : float
            normalized mutual information.
        ari : float
            adjusted Rand index.
        """
        y_pred = torch.argmax(self.q, dim=1).data.cpu().numpy()
        acc = np.round(cluster_acc(y, y_pred), 5)
        nmi = np.round(metrics.normalized_mutual_info_score(y, y_pred), 5)
        ari = np.round(metrics.adjusted_rand_score(y, y_pred), 5)
        return acc, nmi, ari


class MeanAct(nn.Module):
    """Mean activation class.
    """

    def __init__(self):
        super(MeanAct, self).__init__()

    def forward(self, x):
        return torch.clamp(torch.exp(x), min=1e-5, max=1e6)


class DispAct(nn.Module):
    """Dispersion activation class.
    """

    def __init__(self):
        super(DispAct, self).__init__()

    def forward(self, x):
        return torch.clamp(F.softplus(x), min=1e-4, max=1e4)


# Copyright 2022 DSE lab.  All rights reserved.

# TODO：start from project root directory to solve the package import error
import sys
sys.path.append("../../..")
from dance.datasets.spatial import SpotDataset
from dance.modules.spatial.spatial_domain import louvain
from dance.modules.spatial.spatial_domain.spagcn import SpaGCN
from dance.transforms.graph_construct import construct_graph, neighbors_get_adj
from dance.transforms.preprocess import prefilter_cells, prefilter_genes, pca, prefilter_specialgenes, normalize, log1p, set_seed
import argparse

if __name__ == '__main__':
	parser = argparse.ArgumentParser()
	parser.add_argument("--sample_number", type=str, default="151673", \
		help="12 samples of human dorsolateral prefrontal cortex dataset \
		supported in the task of spatial domain task.")
	parser.add_argument("--beta", type=int, default=49, help="")
	parser.add_argument("--alpha", type=int, default=10, help="")
	parser.add_argument("--n_components", type=int, default=50, help="the number of components in PCA")
	parser.add_argument("--neighbors", type=int, default=17, help="")
	args = parser.parse_args()

	set_seed(2020)
	# get data 
	dataset = SpotDataset(args.sample_number, data_dir="../../../data/spot")
	## dataset.data has repeat name , be careful

	# build graph 
	dataset.adj = construct_graph(dataset.data.obs['x'], dataset.data.obs['y'], dataset.data.obs['x_pixel'], dataset.data.obs['y_pixel'], dataset.img, beta=args.beta, alpha=args.alpha, histology=True)

	# preprocess data 
	dataset.data.var_names_make_unique()
	
	# prefilter_cells(dataset.data) # this operation will change the data shape
	prefilter_specialgenes(dataset.data)
	normalize(dataset.data)
	log1p(dataset.data)
 
	pca(dataset.data, n_components=args.n_components)
	dataset.adj = neighbors_get_adj(dataset.data, n_neighbors=args.neighbors, use_rep='X_pca')
	model = louvain.Louvain()
	
	model.fit(dataset.data, dataset.adj, resolution=1)
	prediction = model.predict()
	# print("prediction:", prediction)
	print(model.score(dataset.data.obs['label'].values)) # 0.3058 # 0.33 target


# python louvain.py --beta=49 --alpha=10 --neighbors=25 --n_components=50    0.3058


"""
This is the implementation of DeepImpute.
Extended from https://github.com/lanagarmire/DeepImpute

Reference
----------
Arisdakessian, Cédric, et al. "DeepImpute: an accurate, fast, and scalable deep neural network method to impute single-cell RNA-seq data." Genome biology 20.1 (2019): 1-14.
"""

import pandas as pd
import torch
import torch.nn as nn
import torch.nn.functional as F
import tempfile
import numpy as np
from math import floor
from torch.utils.data import Dataset, DataLoader
import torch.optim as optim
from pathlib import Path
import anndata as ad
import scanpy as sc
from sklearn.metrics import adjusted_rand_score


class NeuralNetworkModel(nn.Module):

    """ model class.
        Parameters
        ----------
        None
        Returns
        -------
        None
    """    
    def __init__(self, inputdim,sub_outputdim, hidden_dim=None,dropout=0.2):
        super(NeuralNetworkModel, self).__init__()
        if(hidden_dim is None):
            hidden_dim = floor(sub_outputdim/2)
        print("hidden_dim: %f" % (hidden_dim))
        self.layer1 = nn.Linear(inputdim, hidden_dim)
        self.layer2 = nn.Dropout(p=dropout)
        self.layer3 = nn.Linear(hidden_dim, sub_outputdim)


    def forward(self, x):
        x=F.relu(self.layer1(x))
        x=self.layer2(x)
        x=F.softplus(self.layer3(x))
        return(x)


class DeepImpute():

    """DeepImpute class.
    Parameters
    ----------
    dl_params : 
        parameters including input information
    learning_rate : float optional
        learning rate
    batch_size : int optional
        batch size
    max_epochs : int optional
        maximum epochs
        
    patience : int optional
        number of epochs before stopping once loss stops to improve
    gpu : int optional
        option to use gpu
    loss : string optional
        loss function
    output_prefix : string optinal
        directory to save outputs
    sub_outputdim : int optional
        output dimensions in each subnetwork
    hidden_dim : int optional
        dimension of the dense layer in each subnetwork
    
    verbose: int optional
        verbose option
    
    seed: int optional
        random seed
    architecture: optional
        network architecture
    
    imputed_only: boolean optional
        whether to return imputed genes only 
    
    policy: string optional
        imputation policy
    """

    def __init__(self,
                 dl_params,
                 learning_rate=1e-5,
                 batch_size=64,
                 max_epochs=500,
                 patience=5,
                 gpu=-1,
                 loss="wMSE",
                 output_prefix=tempfile.mkdtemp(),
                 sub_outputdim=512,
                 hidden_dim=None,
                 verbose=1,
                 seed=1234,
                 architecture=None,
                 imputed_only=False,
                 policy = 'restore'
                 ):
        self.NN_parameters = {"learning_rate": learning_rate,
                              "batch_size": batch_size,
                              "loss": loss,
                              "architecture": architecture,
                              "max_epochs": max_epochs,
                              "patience": patience}
        self.gpu = gpu
        self.sub_outputdim = sub_outputdim
        self.hidden_dim = hidden_dim
        self.outputdir = output_prefix
        self.verbose = verbose
        self.imputed_only = imputed_only
        self.policy = policy
        self.seed = seed
        self.batch_size = batch_size
        self.X_train = dl_params.X_train
        self.Y_train = dl_params.Y_train
        self.X_test = dl_params.X_test
        self.Y_test = dl_params.Y_test
        self.targetgenes = dl_params.targetgenes
        self.inputgenes = dl_params.inputgenes
        self.total_counts = dl_params.total_counts
        self.true_counts=dl_params.true_counts
        self.genes_to_impute = dl_params.genes_to_impute
        self.prj_path = Path(__file__).parent.resolve().parent.resolve().parent.resolve().parent.resolve().parent.resolve()
        self.save_path = self.prj_path / 'example' / 'single_modality' / 'imputation' / 'pretrained' / dl_params.train_dataset / 'models'
        if not self.save_path.exists():
            self.save_path.mkdir(parents=True)
        self.dl_params = dl_params
        self.device = torch.device('cuda' if self.gpu != -1 and torch.cuda.is_available() else 'cpu')



    def loadDefaultArchitecture(self):

        """load default model architecture.
        Parameters
        ----------
        None
        Returns
        -------
        None
        """
        
        if(self.hidden_dim is None):
            hidden_dim =self.sub_outputdim//2
        else:
            hidden_dim = self.hidden_dim
        self.NN_parameters['architecture'] = [
                {"type": "dense", "neurons": hidden_dim, "activation": "relu"},
                {"type": "dropout", "rate": 0.2},
            ]

    def wMSE(self, y_true, y_pred, binary=False):

        """weighted MSE.
        Parameters
        ----------
        y_true: array
            true expression
        Y_train: array
            predicted expression
        binary: boolean optional
            whether to use binary weights
        Returns
        -------
        val: float
            weighted MSE
        """   

        if binary:
            tmp = y_true > 0
            weights = tmp.type(torch.FloatTensor)
        else:
            weights = y_true
        val = torch.mean(weights * torch.square(y_true - y_pred))
        return val

    def build(self, inputdims):
        """build model.
        Parameters
        ----------
        inputdims: int 
            number of neurons as input in the first layer
        Returns
        -------
        models : array
            array of subnetworks
        """

        if self.NN_parameters['architecture'] is None:
            self.loadDefaultArchitecture()

        print(self.NN_parameters['architecture'])

        models = []
        for dim in inputdims:
            models.append(NeuralNetworkModel(dim, self.sub_outputdim, hidden_dim = self.hidden_dim,dropout=self.NN_parameters['architecture'][1]['rate']))

        return models


    def fit(self, X_train=None, Y_train = None,X_test=None,Y_test=None,inputgenes=None):

        """Train model.
        Parameters
        ----------
        X_train: optional
            Training data including input genes
        Y_train: optional
            Training data including target genes to be inputed
        X_test:  optional
            Validation data including input predictor genes
        Y_test:  optional
            Validation data including target genes to be inputed
        inputgenes: array optional
            input genes as predictors for target genes
        Returns
        -------
        None
        """

        if(X_train is None):
            X_train = self.X_train
        if(Y_train is None):
            Y_train = self.Y_train
        if(X_test is None):
            X_test = self.X_test
        if(Y_test is None):
            Y_test = self.Y_test
        if(inputgenes is None):
            inputgenes = self.inputgenes

        device = self.device
        print("Building network")
        self.models = self.build([len(genes) for genes in inputgenes])


        data = [torch.cat((torch.Tensor(X_train[i]), torch.Tensor(Y_train[i])), dim=1) for i in range(len(X_train))]
        train_loaders = [DataLoader(data[i], batch_size=self.batch_size, shuffle=True) for i in range(len(data))]
        for i, model in enumerate(self.models):
            optimizer = optim.Adam(model.parameters(), lr = self.NN_parameters['learning_rate'])
            val_losses = []
            for epoch in range(self.NN_parameters['max_epochs']):
                epoch_loss = 0
                train_loader = train_loaders[i]
                model.train()
                for batch_idx, data in enumerate(train_loader):
                    X_batch = data[:, :(data.shape[1] - self.sub_outputdim)].to(device)
                    Y_batch = data[:,(data.shape[1] - self.sub_outputdim): ].to(device)
                    model.to(device)
                    y_pred = model.forward(X_batch)
                    loss = self.wMSE(Y_batch, y_pred)
                    loss.backward()
                    optimizer.step()
                    epoch_loss += loss.item()

                average_epoch_loss = epoch_loss / len(train_loader)
                X_test_i = torch.Tensor(X_test[i])
                Y_test_i = torch.Tensor(Y_test[i])
                val_loss, _ = self.eval(X_test_i, Y_test_i, model)
                val_losses.append(val_loss.item())
                min_val = min(val_losses)
                patience = self.NN_parameters['patience']
                if val_loss == min_val:
                    #self.save_model(model, optimizer, i)
                    self.models[i] = model
                    print('Saving model %s.' % i)
                print("Average epoch loss: %f, epoch eval loss: %f." % (average_epoch_loss, val_loss))
                # if epoch >= patience and min(val_losses[-patience:]) > min_val:
                #     print("Early stopping on epoch %s." % epoch)
                #     model = self.load_model(model, i)
                #     break
            #self.models[i] = model


    def save_model(self, model, optimizer, i):

        """save model.
        Parameters
        ----------
        model: 
            model to be saved
        
        optimizer:  
            optimizer
        
        i: int
            index of the subnetwork to be loaded
        Returns
        -------
        None
        """

        model_string = 'model_' + str(i)
        opt_string = 'optimizer_' + str(i)
        state = {
            model_string: model.state_dict(),
            opt_string: optimizer.state_dict()
        }
        torch.save(state, self.save_path / f"{self.dl_params.train_dataset}_{i}.pt")

    def load_model(self, model, i):

        """load model.
        Parameters
        ----------
        model: 
            model to be loaded
        
        i: int
            index of the subnetwork to be loaded
        Returns
        -------
        model : 
            loaded model
        """

        model_path = self.save_path / f"{self.dl_params.train_dataset}_{i}.pt"
        state = torch.load(model_path, map_location=self.device)
        model_string = 'model_' + str(i)
        model.load_state_dict(state[model_string])
        return model

    def predict(self, targetgenes=None):

        """Get predictions from the trained model.
        Parameters
        ----------
        targetgenes: array optional
            genes to be imputed
        Returns
        -------
        imputed : DataFrame
            imputed gene expression
        """

        if(targetgenes is None):
            targetgenes = self.targetgenes

        device = self.device

        norm_data = np.log1p(self.total_counts)  # is norm_raw a saved variable?

        # select input data
        X_data = [torch.Tensor(norm_data.loc[:, inputgenes].values).to(device) for inputgenes in self.inputgenes]
        # inputs = [norm_raw.loc[:, predictors].values.astype(np.float32)
        #          for predictors in self.inputgenes]
        # model = self.load_model()
        # predicted = model.predict(inputs)

        # make predictions using each subnetwork
        Y_pred_lst = []
        for i, model in enumerate(self.models):
            model.eval()
            with torch.no_grad():
                Y_pred_lst.append(model.forward(X_data[i]))
        # concatenate predicted values
        Y_pred = torch.cat(Y_pred_lst, dim=1)
        # convert to pd
        Y_pred = pd.DataFrame(Y_pred.cpu().numpy(), index=self.total_counts.index, columns=targetgenes.flatten())

        # create imputed matrix
        predicted = Y_pred.groupby(by=Y_pred.columns, axis=1).mean()
        not_predicted = norm_data.drop(targetgenes.flatten(), axis=1)
        imputed = (pd.concat([predicted, not_predicted], axis=1)
                   .loc[self.total_counts.index, self.total_counts.columns]
                   .values)

        # To prevent overflow
        imputed[(imputed > 2 * norm_data.values.max()) | (np.isnan(imputed))] = 0
        # Convert back to counts
        imputed = np.expm1(imputed)

        if self.policy == "restore":
            print("Filling zeros")
            mask = (self.total_counts.values > 0)
            imputed[mask] = self.total_counts.values[mask]
        elif self.policy == "max":
            print("Imputing data with 'max' policy")
            mask = (self.total_counts.values > imputed)
            imputed[mask] = self.total_counts.values[mask]

        imputed = pd.DataFrame(imputed, index=self.total_counts.index, columns=self.total_counts.columns)

        if self.imputed_only:
            return imputed.loc[:, predicted.columns]
        else:
            return imputed

     



    def score(self,true_expr,metric,true_labels=None,n_neighbors=None,n_pcs=None,clu_resolution=1,targetgenes=None):

        """Evaluate the trained model.
        Parameters
        ----------
        true_expr : DataFrame
            true expression
        metric : string
            choice of metric, either MSE or ARI
        true_labels : array optional
            provided cell labels
        n_neighbors : int optional
            number of neighbors tp cluster imputed data
        n_pcs: int optional
            number of principal components for neighbor detection
        clu_resolution : int optional
            resolution for Leiden cluster
        targetgenes: array optional
            genes to be imputed
        Returns
        -------
        ari : float
            adjusted Rand index.
        MSE_cells : 
            mean squared errors averaged across genes per cell
        MSE_genes : 
            mean squared errors averaged across cellss per gene
        """

        if(targetgenes==None): # set target genes
            targetgenes = self.targetgenes
        imputed_expr = self.predict(targetgenes=targetgenes) # prediction
        allowd_metrics = {"MSE","ARI"}
        if metric not in allowd_metrics:
            raise ValueError("scoring metric %r." % allowd_metrics)
        # subset target genes only
        targetgenes = targetgenes.flatten() # flatten target genes list
        imputed_expr = imputed_expr.loc[:,targetgenes] 
        imputed_expr=imputed_expr.groupby(by=imputed_expr.columns, axis=1).mean() 
        if(metric=='MSE'):
            true_expr = self.true_counts
            masked_expr = self.total_counts
            mask = (masked_expr!= true_expr)
            df_deepImpute =np.log1p(imputed_expr) 
            gene_subset = df_deepImpute.columns
            true_expr = np.log1p(true_expr.reindex(columns=gene_subset))
            masked_expr = np.log1p(masked_expr.reindex(columns=gene_subset))
            mask = mask.reindex(columns=gene_subset)
            MSE_cells=pd.DataFrame(((df_deepImpute[mask]-true_expr[mask])**2).mean(axis=1)).dropna()
            MSE_genes = pd.DataFrame(((df_deepImpute[mask]-true_expr[mask])**2).mean(axis=0)).dropna()
            return MSE_cells, MSE_genes 
        elif(metric=='ARI'):
            adata = ad.AnnData(imputed_expr)
            sc.pp.normalize_total(adata, target_sum=1e4)
            sc.pp.log1p(adata)
            sc.pp.highly_variable_genes(adata)
            adata = adata[:, adata.var.highly_variable]
            sc.pp.scale(adata)
            sc.tl.pca(adata, svd_solver='arpack')
            sc.pp.neighbors(adata, n_neighbors=n_neighbors, n_pcs=n_pcs)
            sc.tl.umap(adata)
            sc.tl.leiden(adata,resolution = clu_resolution)
            pred_labels=adata.obs['leiden']
            ari=round(adjusted_rand_score(true_labels, pred_labels), 3)
            return ari




    def eval(self, X_test, Y_test, model):
        
        """evaluate model.
        Parameters
        ----------
        X_test:  
            Validation data including input predictor genes
        Y_test: 
            Validation data including target genes to be inputed
        
        model: 
            model to be evaluated
        Returns
        -------
        loss : float
            weighted MSE as loss
        y_pred:
            predicted expression
        """
        device = torch.device('cuda' if self.gpu != -1 and torch.cuda.is_available() else 'cpu')
        model.eval()
        with torch.no_grad():
            y_pred = model.forward(X_test.to(device))
            loss = self.wMSE(Y_test.to(device), y_pred)
        return loss, y_pred

 


# Copyright 2022 DSE lab.  All rights reserved.
import math
import os
import random
import numpy as np
import dgl
import torch
from torch.utils.data import Dataset

class SimpleIndexDataset(Dataset):
    def __init__(self,dataset):
        self.data = dataset

    def __len__(self):
        return len(self.data)

    def __getitem__(self, index):
        x = self.data[index]
        return x

class PairedDataset(Dataset):
    """
    Combines two datasets into one, where input is now (x1, x2) and
    output is (y1, y2). A Paired dataset simply combines x and y
    by returning the x input and y input as a tuple, and the x output
    and y output as a tuple, and does not "cross" between the datasets
    """

    # Inherits the init from SplicedDataset since we're doing the same thing - recording
    # the two different datasets
    def __init__(self, dataset_x, dataset_y):
        self.dataset_x = dataset_x
        self.dataset_y = dataset_y

    def __len__(self):
        return len(self.dataset_x)

    def __getitem__(self, i):
        x1 = self.dataset_x[i]
        x2 = self.dataset_y[i]
        x_pair = (x1, x2)
        y_pair = (x1, x2)
        return torch.cat(x_pair), torch.cat(y_pair)

def set_seed(rndseed, cuda=True, extreme_mode = False):
    os.environ['PYTHONHASHSEED'] = str(rndseed)
    random.seed(rndseed)
    np.random.seed(rndseed)
    torch.manual_seed(rndseed)
    if cuda:
        torch.cuda.manual_seed(rndseed)
        torch.cuda.manual_seed_all(rndseed)
    if extreme_mode:
        torch.backends.cudnn.benchmark = False
        torch.backends.cudnn.deterministic = True
    dgl.seed(rndseed)
    dgl.random.seed(rndseed)
    print('Global random seed:', rndseed)

def calculate_p(adj, l):
    adj_exp=np.exp(-1*(adj**2)/(2*(l**2)))
    return np.mean(np.sum(adj_exp,1))-1

def search_l(p, adj, start=0.01, end=1000, tol=0.01, max_run=100):
    run=0
    p_low=calculate_p(adj, start)
    p_high=calculate_p(adj, end)
    if p_low>p+tol:
        print("l not found, try smaller start point.")
        return None
    elif p_high<p-tol:
        print("l not found, try bigger end point.")
        return None
    elif  np.abs(p_low-p) <=tol:
        print("recommended l = ", str(start))
        return start
    elif  np.abs(p_high-p) <=tol:
        print("recommended l = ", str(end))
        return end
    while (p_low+tol)<p<(p_high-tol):
        run+=1
        print("Run "+str(run)+": l ["+str(start)+", "+str(end)+"], p ["+str(p_low)+", "+str(p_high)+"]")
        if run >max_run:
            print("Exact l not found, closest values are:\n"+"l="+str(start)+": "+"p="+str(p_low)+"\nl="+str(end)+": "+"p="+str(p_high))
            return None
        mid=(start+end)/2
        p_mid=calculate_p(adj, mid)
        if np.abs(p_mid-p)<=tol:
            print("recommended l = ", str(mid))
            return mid
        if p_mid<=p:
            start=mid
            p_low=p_mid
        else:
            end=mid
            p_high=p_mid
    return None



# Copyright 2022 DSE lab.  All rights reserved.

# TODO：start from project root directory to solve the package import error

from sklearn.metrics import adjusted_mutual_info_score
from dance.datasets.spatial import SpotDataset
from dance.modules.spatial.spatial_domain.spagcn import SpaGCN, refine
from dance.transforms.graph_construct import construct_graph
from dance.transforms.preprocess import prefilter_cells, prefilter_genes, prefilter_specialgenes, normalize, log1p
import argparse
import sys
sys.path.append("../../..")

if __name__ == '__main__':
	parser = argparse.ArgumentParser()
	parser.add_argument("--sample_number", type=str, default="151673", \
		help="12 samples of human dorsolateral prefrontal cortex dataset \
		supported in the task of spatial domain task.")
	parser.add_argument("--beta", type=int, default=49, help="")
	parser.add_argument("--alpha", type=int, default=1, help="")
	parser.add_argument("--p", type=float, default=0.05, help="percentage of total expression contributed by neighborhoods.")
	parser.add_argument("--l", type=float, default=0.5, help="the parameter to control percentage p.")
	parser.add_argument("--start", type=float, default=0.01, help="starting value for searching l.")
	parser.add_argument("--end", type=float, default=1000, help="ending value for searching l.")
	parser.add_argument("--tol", type=float, default=5e-3, help="tolerant value for searching l.")
	parser.add_argument("--max_run", type=int, default=200, help="max runs.")
	parser.add_argument("--max_epochs", type=int, default=200, help="max epochs.")
	parser.add_argument("--n_clusters", type=int, default=7, help="the number of clusters")
	parser.add_argument("--step", type=float, default=0.1, help="")
	parser.add_argument("--lr", type=float, default=0.05, help="learning rate")
	parser.add_argument("--r_seed", type=int, default=100, help="")
	parser.add_argument("--t_seed", type=int, default=100, help="")
	parser.add_argument("--n_seed", type=int, default=100, help="")
	args = parser.parse_args()


	# get data 
	dataset = SpotDataset(args.sample_number, data_dir="../../../data/spot")
	## dataset.data has repeat name , be careful


	# build graph 
	dataset.adj = construct_graph(dataset.data.obs['x'], dataset.data.obs['y'], dataset.data.obs['x_pixel'], dataset.data.obs['y_pixel'], dataset.img, beta=args.beta, alpha=args.alpha, histology=True)

	# preprocess data 
	dataset.data.var_names_make_unique()

	# prefilter_cells(dataset.data) # this operation will change the data shape
	prefilter_specialgenes(dataset.data)
	normalize(dataset.data)
	log1p(dataset.data)


	# model and pipeline 
	model = SpaGCN()
	
	p = args.p
	l = 0.5
	l = model.search_l(p, dataset.adj, start=args.start, end=args.end, tol=args.tol, max_run=args.max_run)
	model.set_l(0.05)
	n_clusters=args.n_clusters
	res = model.search_set_res(dataset.data, dataset.adj, l=l, target_num=n_clusters, start=0.4, step=args.step, tol=args.tol, lr=args.lr, max_epochs=args.max_epochs, r_seed=args.r_seed, t_seed=args.t_seed, n_seed=args.n_seed, max_run=args.max_run)
	model.fit(dataset.data, dataset.adj, init_spa=True, init="louvain", tol=args.tol, lr=args.lr, max_epochs=args.max_epochs, res=res)
	predict = model.predict()
	adj_2d = construct_graph(x = dataset.data.obs['x_pixel'], y = dataset.data.obs['y_pixel'], histology=False)
	refined_pred = refine(sample_id = dataset.data.obs.index.tolist(), pred= predict[0].tolist(), dis= adj_2d, shape='hexagon')
	print(model.score(dataset.data.obs['label'].values))
	print(adjusted_mutual_info_score(dataset.data.obs['label'].values, refined_pred)) 

""" To reproduce SpaGCN on other samples, please refer to command lines belows:
human dorsolateral prefrontal cortex sample 151676:
python spagcn.py --sample_number=151676 

human dorsolateral prefrontal cortex sample 151507:
python spagcn.py --sample_number=151507 

"""
